import warnings
from collections import defaultdict
from itertools import product
from typing import List, Optional

from lark import Lark

from pharmpy.model import Model
from pharmpy.modeling.covariate_effect import get_covariates
from pharmpy.modeling.odes import (
    get_number_of_peripheral_compartments,
    get_number_of_transit_compartments,
    has_first_order_absorption,
    has_first_order_elimination,
    has_instantaneous_absorption,
    has_lag_time,
    has_michaelis_menten_elimination,
    has_mixed_mm_fo_elimination,
    has_seq_zo_fo_absorption,
    has_zero_order_absorption,
    has_zero_order_elimination,
)
from pharmpy.tools.mfl.feature.covariate import features as covariate_features
from pharmpy.tools.mfl.statement.definition import Let
from pharmpy.tools.mfl.statement.feature.absorption import Absorption
from pharmpy.tools.mfl.statement.feature.covariate import Covariate
from pharmpy.tools.mfl.statement.feature.elimination import Elimination
from pharmpy.tools.mfl.statement.feature.lagtime import LagTime
from pharmpy.tools.mfl.statement.feature.peripherals import Peripherals
from pharmpy.tools.mfl.statement.feature.symbols import Name
from pharmpy.tools.mfl.statement.feature.transits import Transits

from .grammar import grammar
from .helpers import all_funcs, funcs, modelsearch_features
from .interpreter import MFLInterpreter
from .statement.feature.covariate import Ref
from .statement.feature.symbols import Wildcard
from .statement.statement import Statement
from .stringify import stringify as mfl_stringify


def parse(code: str, mfl_class=False) -> List[Statement]:
    parser = Lark(
        grammar,
        start='start',
        parser='lalr',
        # lexer='standard',  # NOTE: This does not work because lexing for the
        #      MFL grammar is context-dependent
        propagate_positions=False,
        maybe_placeholders=False,
        debug=False,
        cache=True,
    )

    tree = parser.parse(code)

    mfl_statement_list = MFLInterpreter().interpret(tree)
    validate_mfl_list(mfl_statement_list)

    # TODO : only return class once it has been implemented everywhere
    if mfl_class:
        return ModelFeatures.create_from_mfl_statement_list(mfl_statement_list)
    else:
        return mfl_statement_list


def validate_mfl_list(mfl_statement_list):
    # TODO : Implement for other features as necessary
    optional_cov = set()
    mandatory_cov = set()
    # FIXME (?) : Allow for same exact cov effect to be forced by multiple explicit statements
    for s in mfl_statement_list:
        if isinstance(s, Covariate):
            if not s.optional.option and isinstance(s.fp, Wildcard):
                raise ValueError(
                    f"Error in {mfl_stringify([s])} :"
                    f" Mandatory effects need to be explicit (not '*')"
                )
            if not isinstance(s.parameter, Ref) and not isinstance(s.covariate, Ref):
                if s.optional.option:
                    optional_cov.update(product(s.parameter, s.covariate))
                else:
                    if error := [
                        e for e in product(s.parameter, s.covariate) if e in mandatory_cov
                    ]:
                        raise ValueError(
                            f"Covariate effect(s) {error} is being forced by"
                            f" multiple statements. Please force only once"
                        )
                    mandatory_cov.update(product(s.parameter, s.covariate))
    if error := [op for op in optional_cov if op in mandatory_cov]:
        raise ValueError(
            f"The covariate effect(s) {error} : are defined as both mandatory and optional"
        )


class ModelFeatures:
    def __init__(
        self,
        absorption=Absorption((Name('INST'),)),
        elimination=Elimination((Name('FO'),)),
        transits=(Transits((0,), (Name('DEPOT'),)),),  # NOTE : This is a tuple
        peripherals=Peripherals((0,)),
        lagtime=LagTime((Name('OFF'),)),
        covariate=tuple(),  # Note : Should always be tuple (empty meaning no covariates)
    ):
        self._absorption = absorption
        self._elimination = elimination
        self._transits = transits
        self._peripherals = peripherals
        self._lagtime = lagtime
        self._covariate = covariate

    @classmethod
    def create(
        cls,
        absorption=Absorption((Name('INST'),)),
        elimination=Elimination((Name('FO'),)),
        transits=(Transits((0,), (Name('DEPOT'),)),),
        peripherals=Peripherals((0,)),
        lagtime=LagTime((Name('OFF'),)),
        covariate=tuple(),
    ):
        # TODO : Check if allowed input value
        if not isinstance(absorption, Absorption):
            raise ValueError(f"Absorption : {absorption} is not suppoerted")

        if not isinstance(elimination, Elimination):
            raise ValueError(f"Elimination : {elimination} is not supported")

        if not isinstance(transits, tuple):
            raise ValueError("Transits need to be given within a tuple")
        if not all(isinstance(t, Transits) for t in transits):
            raise ValueError("All given elements of transits must be of type Transits")

        if not isinstance(peripherals, Peripherals):
            raise ValueError(f"Peripherals : {peripherals} is not supported")

        if not isinstance(lagtime, LagTime):
            raise ValueError(f"Lagtime : {lagtime} is not supported")

        if not covariate == tuple():
            if not isinstance(covariate, tuple):
                raise ValueError("Covariates need to be given within a tuple")
            if not all(isinstance(c, Covariate) for c in covariate):
                raise ValueError(f"Covariate : {covariate} is not supported")

        return cls(
            absorption=absorption,
            elimination=elimination,
            transits=transits,
            peripherals=peripherals,
            lagtime=lagtime,
            covariate=covariate,
        )

    @classmethod
    def create_from_mfl_statement_list(cls, mfl_list):
        m = ModelFeatures()
        absorption = m._default_values()['absorption']
        elimination = m._default_values()['elimination']
        transits = m._default_values()['transits']
        peripherals = m._default_values()['peripherals']
        lagtime = m._default_values()['lagtime']
        covariate = m._default_values()['covariate']
        let = {}
        for statement in mfl_list:
            if isinstance(statement, Absorption):
                absorption = statement
            elif isinstance(statement, Elimination):
                elimination = statement
            elif isinstance(statement, Transits):
                transits = (statement,)
            elif isinstance(statement, Peripherals):
                peripherals = statement
            elif isinstance(statement, LagTime):
                lagtime = statement
            elif isinstance(statement, Covariate):
                covariate += (statement,)
            elif isinstance(statement, Let):
                let[statement.name] = statement.value
            else:
                raise ValueError(f'Unknown ({type(statement)} statement ({statement}) given.')

        # Substitute all Let statements (if any)
        if len(let) != 0:
            # FIXME : Multiple let statements for the same reference value ?
            def _let_subs(cov, let):
                return Covariate(
                    parameter=cov.parameter
                    if not (isinstance(cov.parameter, Ref) and cov.parameter.name in let)
                    else let[cov.parameter.name],
                    covariate=cov.covariate
                    if not (isinstance(cov.covariate, Ref) and cov.covariate.name in let)
                    else let[cov.covariate.name],
                    fp=cov.fp,
                    op=cov.op,
                    optional=cov.optional,
                )

            covariate = tuple([_let_subs(cov, let) for cov in covariate])

        mfl = cls.create(
            absorption=absorption,
            elimination=elimination,
            transits=transits,
            peripherals=peripherals,
            lagtime=lagtime,
            covariate=covariate,
        )
        return mfl

    @classmethod
    def create_from_mfl_string(cls, mfl_string):
        return parse(mfl_string, mfl_class=True)

    def replace(self, **kwargs):
        absorption = kwargs.get("absorption", self._absorption)
        elimination = kwargs.get("elimination", self._elimination)
        transits = kwargs.get("transits", self._transits)
        peripherals = kwargs.get("peripherals", self._peripherals)
        lagtime = kwargs.get("lagtime", self._lagtime)
        covariate = kwargs.get("covariate", self._covariate)
        return ModelFeatures.create(
            absorption=absorption,
            elimination=elimination,
            transits=transits,
            peripherals=peripherals,
            lagtime=lagtime,
            covariate=covariate,
        )

    def _default_values(self):
        return {
            'absorption': Absorption((Name('INST'),)),
            'elimination': Elimination((Name('FO'),)),
            'transits': (Transits((0,), (Name('DEPOT'),)),),
            'peripherals': Peripherals((0,)),
            'lagtime': LagTime((Name('OFF'),)),
            'covariate': tuple(),
        }

    @property
    def absorption(self):
        return self._absorption

    @property
    def elimination(self):
        return self._elimination

    @property
    def transits(self):
        return self._transits

    @property
    def peripherals(self):
        return self._peripherals

    @property
    def lagtime(self):
        return self._lagtime

    @property
    def covariate(self):
        return self._covariate

    def expand(self, model):
        explicit_covariates = set(
            [
                p
                for c in self.covariate
                if (not isinstance(c.parameter, Ref) and not isinstance(c.covariate, Ref))
                for p in product(c.parameter, c.covariate)
            ]
        )  # Override @ reference with explicit value
        covariate = tuple(
            c for c in [c.eval(model, explicit_covariates) for c in self.covariate] if c is not None
        )

        param_cov = [
            p for c in covariate for p in product(c.parameter, c.covariate) if not c.optional.option
        ]
        counts = [(c, param_cov.count(c)) for c in param_cov]
        if any(c[1] > 1 for c in counts):
            error = set(c[0] for c in filter(lambda c: c[1] > 1, counts))
            raise ValueError(
                f"Covariate effect(s) {error} is forced by multiple reference statements."
                f" Please redefine the search space."
            )

        return ModelFeatures.create(
            absorption=self.absorption.eval,
            elimination=self.elimination.eval,
            transits=tuple([t.eval for t in self.transits]),
            peripherals=self.peripherals,
            lagtime=self.lagtime.eval,
            covariate=covariate,
        )

    def mfl_statement_list(self, attribute_type: Optional[List[str]] = []):
        """Add the repspective MFL attributes to a list"""

        # NOTE : This function is needed to be able to convert the classes to functions

        if not attribute_type:
            attribute_type = [
                "absorption",
                "elimination",
                "transits",
                "peripherals",
                "lagtime",
                "covariate",
            ]
        mfl_list = []
        if "absorption" in attribute_type:
            mfl_list.append(self.absorption)
        if "elimination" in attribute_type:
            mfl_list.append(self.elimination)
        if "transits" in attribute_type:
            for t in self.transits:
                mfl_list.append(t)
        if "peripherals" in attribute_type:
            mfl_list.append(self.peripherals)
        if "lagtime" in attribute_type:
            mfl_list.append(self.lagtime)
        if "covariate" in attribute_type:
            for c in self.covariate:
                mfl_list.append(c)

        return [m for m in mfl_list if m]

    def convert_to_funcs(
        self, attribute_type: Optional[List[str]] = None, model: Optional[Model] = None
    ):
        # The model argument is used for when extacting covariates.
        if not model:
            model = Model()
        # TODO : implement argument if we wish to use subset instead of all functions.
        if self.covariate != tuple():
            return all_funcs(model, self.mfl_statement_list(attribute_type))
        else:
            return funcs(model, self.mfl_statement_list(attribute_type), modelsearch_features)

    def contain_subset(self, mfl, model: Optional[Model] = None, tool: Optional[str] = None):
        """See if class contain specified subset"""
        transits = self._subset_transits(mfl)

        # FIXME : Add support for wildcard
        if (
            all([s in self.absorption.eval.modes for s in mfl.absorption.eval.modes])
            and all([s in self.elimination.eval.modes for s in mfl.elimination.eval.modes])
            and transits
            and all([s in self.peripherals.counts for s in mfl.peripherals.counts])
            and all([s in self.lagtime.eval.modes for s in mfl.lagtime.eval.modes])
        ):
            if tool is None or tool in ["modelsearch"]:
                return True
            else:
                if self.covariate != tuple() or mfl.covariate != tuple():
                    if model is None:
                        warnings.warn("Need argument 'model' in order to compare covariates")
                    else:
                        return True if self._subset_covariate(mfl, model) else False
        else:
            return False

    def _subset_transits(self, mfl):
        lhs_counts = set([c for t in self.transits for c in t.counts])
        lhs_depot = set([d for t in self.transits for d in t.eval.depot])

        rhs_counts = set([c for t in mfl.transits for c in t.counts])
        rhs_depot = set([d for t in mfl.transits for d in t.eval.depot])
        # FIXME : Need to compare counts per depot individually when comparing two
        # search spaces (Currenty working for model vs search space)
        return all([c in lhs_counts for c in rhs_counts]) and all(
            [d in lhs_depot for d in rhs_depot]
        )

    def _subset_covariates(self, mfl, model):
        lhs = defaultdict(list)
        rhs = defaultdict(list)

        for cov in self.covariate:
            cov_eval = cov.eval(model)
            for effect in cov_eval.fp:
                for op in cov_eval.op:
                    lhs[(effect, op)].append(product(cov_eval.parameter, cov_eval.covariate))
        for cov in mfl.covariate:
            cov_eval = cov.eval(model)
            for effect in cov_eval.fp:
                for op in cov_eval.op:
                    rhs[(effect, op)].append(product(cov_eval.parameter, cov_eval.covariate))

        for key in rhs.keys():
            if key not in lhs.keys():
                return False
            if all(p in lhs[key] for p in rhs[key]):
                continue
            else:
                return False
        return True

    def least_number_of_transformations(
        self, other, model: Optional[Model] = None, tool: Optional[str] = None
    ):
        """The smallest set of transformations to become part of other"""
        # Add more tools than "modelsearch" if support is needed
        lnt = {}
        if tool is None or tool in ["modelsearch"]:
            if not any(a in other.absorption.eval.modes for a in self.absorption.eval.modes):
                name, func = list(other.convert_to_funcs(["absorption"]).items())[0]
                lnt[name] = func

            if not any(e in other.elimination.eval.modes for e in self.elimination.eval.modes):
                name, func = list(other.convert_to_funcs(["elimination"]).items())[0]
                lnt[name] = func

            lnt = self._lnt_transits(other, lnt)

            if not any(p in other.peripherals.counts for p in self.peripherals.counts):
                name, func = list(other.convert_to_funcs(["peripherals"]).items())[0]
                lnt[name] = func

            if not any(lt in other.lagtime.eval.modes for lt in self.lagtime.eval.modes):
                name, func = list(other.convert_to_funcs(["lagtime"]).items())[0]
                lnt[name] = func

        # TODO : Use in covsearch instead of taking diff
        if tool is None:
            if model is not None:
                lnt = self._lnt_covariates(other, lnt, model)
            else:
                if self.covariate != tuple() or other.covariate != tuple():
                    warnings.warn("Need argument 'model' in order to compare covariates")

        return lnt

    def _lnt_transits(self, other, lnt):
        lhs_depot = [d for t in self.transits for d in t.eval.depot]
        rhs_depot = [d for t in other.transits for d in t.eval.depot]

        if not any(td in rhs_depot for td in lhs_depot):
            # DEPOT does not match -> Take first function regardless of counts
            name, func = list(other.convert_to_funcs(["transits"]).items())[0]
            lnt[name] = func
        else:
            # DEPOT does match -> Need to check counts for corresponding depot
            # First check counts of all matching depots
            for depot in lhs_depot:
                if depot in rhs_depot:
                    depot_counts = [
                        c for t in self.transits for c in t.counts if depot in t.eval.depot
                    ]
                    rhs_depot_counts = [
                        c for t in other.transits for c in t.counts if depot in t.eval.depot
                    ]
                    diff = [c for c in rhs_depot_counts if c not in depot_counts]
                    match = [c for c in rhs_depot_counts if c in depot_counts]

                    if len(match) != 0:
                        return lnt
                    # There is a difference in set of counts for the same depot argument
                    if len(diff) != 0:
                        func_dict = other.convert_to_funcs(["transits"])
                        lnt[('TRANSITS', diff[0], depot.name)] = func_dict[
                            ('TRANSITS', diff[0], depot.name)
                        ]

                        return lnt

            # else take first count value of non matching depot
            depot = next(d for d in rhs_depot if d not in lhs_depot)
            count = next(c for t in other.transits for c in t.counts if depot in t.eval.depot)
            func_dict = other.convert_to_funcs(["transits"])
            lnt[('TRANSITS', count, depot.name)] = func_dict[('TRANSITS', diff[0], depot.name)]

        return lnt

    def _lnt_covariates(self, other, lnt, model):
        lhs = self._extract_covariates()
        rhs = other._extract_covariates()

        remove_cov_dict = dict(covariate_features(model, self.covariate, remove=True))
        if not any(key in rhs.keys() for key in lhs.keys()):
            # Remove all covariates
            lnt.update(remove_cov_dict)
            if len(rhs) != 0:
                # No effect/operator combination matching
                func_dict = other.convert_to_funcs(["covariate"])
                key = list(func_dict.keys())[0]
                lnt[key] = func_dict[key]
        else:
            # One (or more) effect/operator are overlapping
            for key in lhs.keys():
                if key in rhs.keys():
                    if not any(p in rhs[key] for p in lhs[key]):
                        # take the first value with this key.
                        func_dict = other.convert_to_funcs(["covariate"], model)
                        func_dict = {
                            k: v
                            for k, v in func_dict.items()
                            if (k[3] == key[0] and k[4] == key[1])
                        }
                        key = list(func_dict.keys())[0]
                        lnt[key] = func_dict[key]
                    for p in [p for p in lhs[key] if p not in rhs[key]]:
                        lnt[(p[0], p[1], key[0], key[1])] = remove_cov_dict[
                            (p[0], p[1], key[0], key[1])
                        ]

        return lnt

    def __repr__(self):
        # TODO : Remove default values
        return mfl_stringify(self.mfl_statement_list())

    def __sub__(self, other):
        transits = self._add_sub_transits(other, add=False)
        covariates = self._add_sub_covariates(other, add=False)

        return ModelFeatures.create(
            absorption=self.absorption - other.absorption,
            elimination=self.elimination - other.elimination,
            transits=transits,
            peripherals=self.peripherals - other.peripherals,
            lagtime=self.lagtime - other.lagtime,
            covariate=covariates,
        )

    def __add__(self, other):
        transits = self._add_sub_transits(other, add=True)
        covariates = self._add_sub_covariates(other, add=True)

        return ModelFeatures.create(
            absorption=self.absorption + other.absorption,
            elimination=self.elimination + other.elimination,
            transits=transits,
            peripherals=self.peripherals + other.peripherals,
            lagtime=self.lagtime + other.lagtime,
            covariate=covariates,
        )

    def _add_sub_transits(self, other, add=True):
        """Apply logic for adding/subtracting mfl transits.
        Use add = False for subtraction"""

        # Need multiple Transit objects due to presence of both depot / nodepot
        # and different sets of counts
        lhs_counts_depot = [
            c for t in self.transits if Name("DEPOT") in t.eval.depot for c in t.counts
        ]
        lhs_counts_nodepot = [
            c for t in self.transits if Name("NODEPOT") in t.eval.depot for c in t.counts
        ]
        rhs_counts_depot = [
            c for t in other.transits if Name("DEPOT") in t.eval.depot for c in t.counts
        ]
        rhs_counts_nodepot = [
            c for t in other.transits if Name("NODEPOT") in t.eval.depot for c in t.counts
        ]

        if add:
            depot_counts = list(set(lhs_counts_depot + rhs_counts_depot))
            nodepot_counts = list(set(lhs_counts_nodepot + rhs_counts_nodepot))
        else:
            depot_counts = [c for c in lhs_counts_depot if c not in rhs_counts_depot]
            nodepot_counts = [c for c in lhs_counts_nodepot if c not in rhs_counts_nodepot]

        if len(depot_counts) == len(nodepot_counts) == 0:
            transits = (
                Transits(
                    (0,),
                    (Name("DEPOT"),),
                ),
            )
        else:
            both_depot = [d for d in depot_counts if d in nodepot_counts]
            nodepot_diff = [d for d in nodepot_counts if d not in depot_counts]
            depot_diff = [d for d in depot_counts if d not in nodepot_counts]

            transits = []
            if len(both_depot) != 0:
                transits.append(Transits(tuple(both_depot), (Name("DEPOT"), Name("NODEPOT"))))
            if len(depot_diff) != 0:
                transits.append(Transits(tuple(depot_diff), (Name("DEPOT"),)))
            if len(nodepot_diff) != 0:
                transits.append(Transits(tuple(nodepot_diff), (Name("NODEPOT"),)))

        return tuple(transits)

    def _add_sub_covariates(self, other, add=True):
        lhs = self._extract_covariates()
        rhs = other._extract_covariates()

        res = []
        if len(rhs) != 0 and len(lhs) != 0:
            # Find the unique products in both lists with matching expression/operator
            combined = lhs
            for effop, prod in rhs.items():
                if add:
                    combined[effop].update(prod)
                else:
                    combined[effop].difference_update(prod)
            combined = {k: v for k, v in combined.items() if v != set()}
            res = combined
        elif len(rhs) != 0 and len(lhs) == 0 and add:
            res = rhs
        elif len(lhs) != 0 and len(rhs) == 0:
            res = lhs
        else:
            return tuple()

        # Clean and combine all effect/operators with the same set of parameter and covariates
        clean_res = defaultdict(list)
        for effop, prod in res.items():
            prod_key = tuple(sorted(prod))
            if existing_effop := clean_res[prod_key]:
                if effop[0] not in existing_effop[0]:
                    existing_effop[0].append(effop[0])
                if effop[1] not in existing_effop[1]:
                    existing_effop[1].append(effop[1])
                clean_res[prod_key] = existing_effop
            else:
                clean_res[prod_key] = [[effop[0]], [effop[1]]]

        cov_res = []
        for prod, effop in clean_res.items():
            parameter = tuple(set([p[0] for p in prod]))
            covariate = tuple(set([p[1] for p in prod]))
            cov_res.append(Covariate(parameter, covariate, tuple(effop[0]), tuple(effop[1])))

        if all(len(c.parameter) == 0 for c in cov_res):
            return tuple()
        else:
            return tuple(cov_res)

    def __eq__(self, other):
        transits = self._eq_transits(other)
        return (
            self.absorption == other.absorption
            and self.elimination == other.elimination
            and transits
            and self.peripherals == other.peripherals
            and self.lagtime == other.lagtime
            and self._eq_covariate(other)
        )

    def _eq_transits(self, other):
        lhs_counts_depot = [
            c for t in self.transits if Name("DEPOT") in t.eval.depot for c in t.counts
        ]
        lhs_counts_nodepot = [
            c for t in self.transits if Name("NODEPOT") in t.eval.depot for c in t.counts
        ]
        rhs_counts_depot = [
            c for t in other.transits if Name("DEPOT") in t.eval.depot for c in t.counts
        ]
        rhs_counts_nodepot = [
            c for t in other.transits if Name("NODEPOT") in t.eval.depot for c in t.counts
        ]
        return set(lhs_counts_depot) == set(rhs_counts_depot) and set(lhs_counts_nodepot) == set(
            rhs_counts_nodepot
        )

    def _eq_covariate(self, other):
        lhs = self._extract_covariates()
        rhs = other._extract_covariates()

        if all(key in rhs.keys() for key in lhs.keys()):
            for key in lhs.keys():
                if lhs[key] != rhs[key]:
                    return False
            return True
        return False

    def _extract_covariates(self):
        lhs = defaultdict(set)
        lhs_ref = []
        for cov in self.covariate:
            if isinstance(cov.parameter, Ref) or isinstance(cov.covariate, Ref):
                lhs_ref.append(cov)
                continue
            for effect in cov.eval().fp:
                for op in cov.op:
                    # TODO : Ignore using PROD and simply use tuple with one value each
                    lhs[(effect.lower(), op)].update(set(product(cov.parameter, cov.covariate)))

        if len(lhs_ref) != 0:
            raise ValueError(
                'Cannot be performed with reference value. Try using .expand(model) first.'
            )

        return lhs


def get_model_features(model: Model, supress_warnings: bool = False) -> str:
    """Create an MFL representation of an input model

    Given an input model. Create a model feature language (MFL) string
    representation. Can currently extract absorption, elimination, transits,
    peripherals and lagtime.

    Parameters
    ----------
    model : Model
        Model to extract features from.
    supress_warnings : TYPE, optional
        Choose to supress warnings if absorption/elimination type cannot be
        determined. The default is False.

    Returns
    -------
    str
        A MFL string representation of the input model.

    """
    # ABSORPTION
    absorption = None
    if has_seq_zo_fo_absorption(model):
        absorption = "SEQ-ZO-FO"
    elif has_zero_order_absorption(model):
        absorption = "ZO"
    elif has_first_order_absorption(model):
        absorption = "FO"
    elif has_instantaneous_absorption(model):
        absorption = "INST"

    if not supress_warnings:
        if absorption is None:
            warnings.warn("Could not determine absorption of model.")

    # ElIMINATION
    elimination = None
    if has_mixed_mm_fo_elimination(model):
        elimination = "MIX-FO-MM"
    elif has_zero_order_elimination(model):
        elimination = "ZO"
    elif has_first_order_elimination(model):
        elimination = "FO"
    elif has_michaelis_menten_elimination(model):
        elimination = "MM"

    if not supress_warnings:
        if elimination is None:
            warnings.warn("Could not determine elimination of model.")

    # ABSORPTION DELAY (TRANSIT AND LAGTIME)
    # TRANSITS
    transits = get_number_of_transit_compartments(model)

    # TODO : DEPOT
    if not model.statements.ode_system.find_depot(model.statements):
        depot = "NODEPOT"
    else:
        depot = "DEPOT"

    lagtime = has_lag_time(model)
    if not lagtime:
        lagtime = None

    # DISTRIBUTION (PERIPHERALS)
    peripherals = get_number_of_peripheral_compartments(model)

    # COVARIATES
    covariates = get_covariates(model)

    if absorption:
        absorption = f'ABSORPTION({absorption})'
    if elimination:
        elimination = f'ELIMINATION({elimination})'
    if lagtime:
        lagtime = "LAGTIME(ON)"
    if transits != 0:
        transits = f'TRANSITS({transits}{","+depot})'
    if peripherals != 0:
        peripherals = f'PERIPHERALS({peripherals})'
    if len(covariates) != 0:
        # FIXME : More extensive cleanup
        clean_cov = defaultdict(list)
        for key, value in covariates.items():
            clean_cov[(key[1], value[0][0], value[0][1])].append(key[0])

        cov_list = [
            f'COVARIATE({value},{key[0]},{key[1]},{key[2]})' for key, value in clean_cov.items()
        ]

        covariates = ';'.join(cov_list)
        # Remove quotes from parameter names
        covariates = covariates.replace("'", '')
    else:
        covariates = None

    # TODO : Implement IIV
    return ";".join(
        [
            e
            for e in [absorption, elimination, lagtime, transits, peripherals, covariates]
            if (e is not None and e != 0)
        ]
    )



from .baseclass import ToolDatabase
from .local_directory import LocalDirectoryToolDatabase
from .null_database import NullToolDatabase

__all__ = [
    'NullToolDatabase',
    'LocalDirectoryToolDatabase',
    'ToolDatabase',
]



import shutil

import pandas as pd
import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import read_model
from pharmpy.tools import fit, run_modelsearch
from pharmpy.workflows import ModelDatabase


def test_exhaustive(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_modelsearch(
            'ABSORPTION([FO,ZO]);PERIPHERALS([0,1])',
            'exhaustive',
            results=start_modelres[1],
            model=start_modelres[0],
        )

        assert len(res.summary_tool) == 4
        assert len(res.summary_models) == 4
        assert len(res.models) == 4

        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == 3
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'metadata.json').exists()


@pytest.mark.parametrize(
    'search_space, no_of_models, last_model_parent_name, model_with_error, ref',
    [
        (
            'ABSORPTION([FO,ZO]);PERIPHERALS([0,1])',
            4,
            'modelsearch_run2',
            'modelsearch_run3',
            ('modelsearch_run2', ['PERIPHERALS(1)', 'VP1 = ']),
        ),
        (
            'ABSORPTION([FO,ZO,SEQ-ZO-FO]);PERIPHERALS([0,1])',
            7,
            'modelsearch_run3',
            'modelsearch_run5',
            ('modelsearch_run3', ['PERIPHERALS(1)', 'VP1 = ']),
        ),
    ],
)
def test_exhaustive_stepwise_basic(
    tmp_path,
    model_count,
    start_modelres,
    search_space,
    no_of_models,
    last_model_parent_name,
    model_with_error,
    ref,
):
    with chdir(tmp_path):
        res = run_modelsearch(
            search_space,
            'exhaustive_stepwise',
            results=start_modelres[1],
            model=start_modelres[0],
        )

        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models + 1

        assert res.models[0].parent_model == 'mox2'
        assert res.models[-1].parent_model == last_model_parent_name
        if last_model_parent_name != 'mox2':
            last_model_features = res.summary_tool.loc[res.models[-1].name]['description']
            parent_model_features = res.summary_tool.loc[last_model_parent_name]['description']
            assert last_model_features[: len(parent_model_features)] == parent_model_features

        if model_with_error:
            assert model_with_error in res.summary_errors.index.get_level_values('model')

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_models
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'metadata.json').exists()

        db: ModelDatabase = res.tool_database.model_database
        model_name, code_ref = ref
        path = db.retrieve_file(model_name, f'{model_name}.mod')
        with open(path, 'r') as fh:
            model_code = fh.read()
            assert all(code in model_code for code in code_ref)


@pytest.mark.parametrize(
    'search_space, iiv_strategy, no_of_models, no_of_added_etas',
    [
        ('ABSORPTION([FO,ZO]);PERIPHERALS([0,1])', 'add_diagonal', 4, 2),
        ('ABSORPTION([FO,ZO]);PERIPHERALS([0,1])', 'fullblock', 4, 2),
        ('ABSORPTION(FO);PERIPHERALS([0,1]);LAGTIME([OFF,ON])', 'absorption_delay', 4, 1),
    ],
)
def test_exhaustive_stepwise_iiv_strategies(
    tmp_path,
    model_count,
    start_modelres,
    search_space,
    iiv_strategy,
    no_of_models,
    no_of_added_etas,
):
    with chdir(tmp_path):
        res = run_modelsearch(
            search_space,
            'exhaustive_stepwise',
            iiv_strategy=iiv_strategy,
            results=start_modelres[1],
            model=start_modelres[0],
        )

        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models + 1
        model_last = res.models[no_of_models - 1]
        assert (
            len(model_last.random_variables.etas.names)
            - len(start_modelres[0].random_variables.etas.names)
            == no_of_added_etas
        )

        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_models
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'metadata.json').exists()


# def test_exhaustive_stepwise_start_model_not_fitted(tmp_path, model_count, start_model):
#    with chdir(tmp_path):
#        start_model = start_model.copy()
#        start_model.name = 'start_model_copy'
#        start_model.modelfit_results = None

#        search_space = 'ABSORPTION(ZO);PERIPHERALS(1)'
#        with pytest.warns(UserWarning, match='Could not update'):
#            res = run_modelsearch(
#                search_space,
#                'exhaustive_stepwise',
#                results=start_model.modelfit_results,
#                model=start_model,
#            )

#        assert len(res.summary_tool) == 5
#        assert len(res.summary_models) == 5
#        assert res.summary_tool['dbic'].isnull().values.all()
#        assert len(res.models) == 4
#        rundir = tmp_path / 'modelsearch_dir1'
#        assert rundir.is_dir()
#        assert model_count(rundir) == 4


def test_exhaustive_stepwise_peripheral_upper_limit(tmp_path, start_modelres):
    with chdir(tmp_path):
        res = run_modelsearch(
            'PERIPHERALS([0,1])',
            'exhaustive_stepwise',
            results=start_modelres[1],
            model=start_modelres[0],
        )

        assert ',999999) ; POP_QP1' in res.models[-1].model_code
        assert ',999999) ; POP_VP1' in res.models[-1].model_code


def test_summary_individuals(tmp_path, testdata):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        m = read_model('pheno_real.mod')
        start_res = fit(m)
        m = m.replace(modelfit_results=start_res)
        res = run_modelsearch(
            model=m,
            results=m.modelfit_results,
            search_space='ABSORPTION(ZO);PERIPHERALS([1, 2])',
            algorithm='reduced_stepwise',
        )
        summary = res.summary_individuals
        columns = (
            'description',
            'parent_model',
            'outlier_count',
            'ofv',
            'dofv_vs_parent',
            'predicted_dofv',
            'predicted_residual',
        )
        assert summary is not None
        assert tuple(summary.columns) == columns
        assert summary['dofv_vs_parent'].equals(
            summary.apply(
                lambda row: summary.loc[(row['parent_model'], row.name[1])]['ofv'] - row['ofv'],
                axis=1,
            )
        )



from dataclasses import dataclass
from typing import Callable, Optional

import pytest

from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools.run import import_tool, run_tool_with_name
from pharmpy.workflows import Results, Task, Workflow, WorkflowBuilder


def test_import_tool():
    from pharmpy.tools import iivsearch

    tool = import_tool('iivsearch')
    assert tool == iivsearch


def create_workflow_rename(new_name, name=None, model: Optional[Model] = None):
    def rename(m):
        m = m.replace(name=new_name)
        return m

    wb = WorkflowBuilder(tasks=[Task('copy', lambda x: x, model)], name=name)
    wb.insert_workflow(WorkflowBuilder(tasks=[Task('rename', rename)]))
    return Workflow(wb)


@with_same_arguments_as(create_workflow_rename)
def validate_input_rename(model, new_name):
    assert isinstance(new_name, str)
    assert isinstance(model, Model)


def create_workflow_generic(name=None, model: Optional[Model] = None):
    return Workflow(WorkflowBuilder(tasks=[Task('copy', lambda _: Results(), model)], name=name))


@with_same_arguments_as(create_workflow_generic)
def validate_input_generic(model):
    assert isinstance(model, Model)


@dataclass(frozen=True)
class MockedTool:
    create_workflow: Callable[..., Workflow]


@dataclass(frozen=True)
class MockedToolWithInputValidation(MockedTool):
    validate_input: Callable[..., None]


@pytest.mark.parametrize(
    ('name', 'tool', 'args', 'expected'),
    (
        ('mocked', MockedTool(create_workflow_generic), (), lambda res: isinstance(res, Results)),
        (
            'mocked',
            MockedToolWithInputValidation(create_workflow_generic, validate_input_generic),
            (),
            lambda res: isinstance(res, Results),
        ),
        (
            'modelfit',  # NOTE: This triggers the modelfit-specific (non-)branches
            MockedTool(create_workflow_rename),
            ('y',),
            lambda res: isinstance(res, Model) and res.name == 'y',
        ),
        (
            'modelfit',  # NOTE: This triggers the modelfit-specific (non-)branches
            MockedToolWithInputValidation(create_workflow_rename, validate_input_rename),
            ('y',),
            lambda res: isinstance(res, Model) and res.name == 'y',
        ),
    ),
)
def test_run_tool_without_input_validation(tmp_path, pheno, name, tool, args, expected):
    with chdir(tmp_path):
        res = run_tool_with_name(name, tool, args, {'name': name, 'model': pheno})
        assert expected(res)



import warnings
from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Literal, Optional, Union

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
else:
    from pharmpy.deps import numpy as np
    import pandas as pd

from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import (
    calculate_parameters_from_ucp,
    calculate_ucp_scale,
    create_rng,
    sample_parameters_uniformly,
    update_inits,
)
from pharmpy.tools import summarize_modelfit_results
from pharmpy.tools.common import ToolResults, create_results
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

SCALES = frozenset(('normal', 'UCP'))


@dataclass
class Retry:
    modelentry: ModelEntry
    number_of_retries: int


def create_workflow(
    model: Optional[Model] = None,
    results: Optional[ModelfitResults] = None,
    number_of_candidates: int = 5,
    fraction: float = 0.1,
    use_initial_estimates: bool = False,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs >= 0.1)",
    scale: Optional[Literal[tuple(SCALES)]] = "UCP",
    prefix_name: Optional[str] = "",  # FIXME : Remove once new database has been implemented
    seed: Optional[Union[np.random.Generator, int]] = None,
):
    """
    Run retries tool.

    Parameters
    ----------
    model : Optional[Model], optional
        Model object to run retries on. The default is None.
    results : Optional[ModelfitResults], optional
        Connected ModelfitResults object. The default is None.
    number_of_candidates : int, optional
        Number of retry candidates to run. The default is 5.
    fraction: float
        Determines allowed increase/decrease from initial parameter estimate. Default is 0.1 (10%)
    use_initial_estimates : bool
        Use initial parameter estimates instead of final estimates of input model when creating candidate models.
    strictness : Optional[str], optional
        Strictness criteria. The default is "minimization_successful or (rounding_errors and sigdigs >= 0.1)".
    scale : {'normal', 'UCP'}
        Which scale to update the initial values on. Either normal scale or UCP scale.
    prefix_name: Optional[str]
        Prefix the candidate model names with given string.
    seed: int or rng
        Random number generator or seed to be used.

    Returns
    -------
    RetriesResults
        Retries tool results object.

    """

    wb = WorkflowBuilder(name='retries')

    if model is not None:
        start_task = Task('Start_retries', _start, results, model)
    else:
        # Remove?
        start_task = Task('Start_retries', _start)

    wb.add_task(start_task)

    candidate_tasks = []
    seed = create_rng(seed)
    for i in range(1, number_of_candidates + 1):
        new_candidate_task = Task(
            f"Create_candidate_{i}",
            create_random_init_model,
            i,
            scale,
            fraction,
            use_initial_estimates,
            prefix_name,
            seed,
        )
        wb.add_task(new_candidate_task, predecessors=start_task)
        candidate_tasks.append(new_candidate_task)
    task_gather = Task('Gather', lambda *retries: retries)
    wb.add_task(task_gather, predecessors=[start_task] + candidate_tasks)

    results_task = Task('Results', task_results, strictness)
    wb.add_task(results_task, predecessors=task_gather)

    return Workflow(wb)


def _start(results, model):
    # Convert to modelentry
    return ModelEntry.create(model=model, modelfit_results=results)


def create_random_init_model(
    context, index, scale, fraction, use_initial_estimates, prefix_name, seed, modelentry
):
    original_model = modelentry.model
    # Update inits once before running
    if not use_initial_estimates and modelentry.modelfit_results:
        original_model = update_inits(
            original_model, modelentry.modelfit_results.parameter_estimates
        )

    # Add any description?
    if prefix_name:
        name = f'{prefix_name}_retries_run{index}'
    else:
        name = f'retries_run{index}'
    new_candidate_model = original_model.replace(name=name)

    try_number = 1
    if scale == "normal":
        maximum_tests = 20  # TODO : Convert to argument

        for try_number in range(1, maximum_tests + 1):
            new_parameters = create_new_parameter_inits(new_candidate_model, fraction, scale, seed)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "error",
                    message="Adjusting initial estimates to create positive semidefinite omega/sigma matrices",
                    category=UserWarning,
                )
                try:
                    new_candidate_model = update_inits(new_candidate_model, new_parameters)
                    break
                except UserWarning:
                    if try_number == maximum_tests:
                        raise ValueError(
                            f"{new_candidate_model.name} could not be determined"
                            f" to be positive semi-definite."
                        )
    elif scale == "UCP":
        new_parameters = create_new_parameter_inits(new_candidate_model, fraction, scale, seed)
        new_candidate_model = update_inits(new_candidate_model, new_parameters)
    else:
        # Should be caught in validate_input()
        raise ValueError(f'Scale ({scale}) is not supported')

    new_candidate_model_entry = ModelEntry.create(
        model=new_candidate_model,
        modelfit_results=None,
        parent=original_model,
    )
    new_candidate_model_fit_wf = create_fit_workflow(modelentries=[new_candidate_model_entry])
    new_modelentry = call_workflow(new_candidate_model_fit_wf, f'fit_candidate_run{index}', context)
    return Retry(
        modelentry=new_modelentry,
        number_of_retries=try_number,
    )


def create_new_parameter_inits(model, fraction, scale, seed):
    if scale == "normal":
        new_parameters = sample_parameters_uniformly(
            model,
            pd.Series(model.parameters.inits),
            fraction=fraction,
            scale=scale,
            seed=seed,
        )
        new_parameters = {p: new_parameters[p][0] for p in new_parameters}
    elif scale == "UCP":
        ucp_scale = calculate_ucp_scale(model)
        new_parameters = sample_parameters_uniformly(
            model,
            pd.Series(model.parameters.inits),
            fraction=fraction,
            scale=scale,
            seed=seed,
        )
        new_parameters = {p: new_parameters[p][0] for p in new_parameters}
        new_parameters = calculate_parameters_from_ucp(model, ucp_scale, new_parameters)
    else:
        raise ValueError(f'{scale} is not supported')
    return new_parameters


def task_results(strictness, retries):
    # Note : the input (modelentry) is a part of retries
    retry_runs = []
    for r in retries:
        if isinstance(r, ModelEntry):
            input_model_entry = r
        elif isinstance(r, Retry):
            retry_runs.append(r)
        else:
            raise ValueError(f'Unknown type ({type(r)}) found when summarizing results.')
    res_models = [r.modelentry for r in retry_runs]
    results_to_summarize = [input_model_entry.modelfit_results] + [
        r.modelentry.modelfit_results for r in retry_runs
    ]
    rank_type = "ofv"
    cutoff = None

    summary_models = summarize_modelfit_results(results_to_summarize)
    summary_models['step'] = [0] + [1] * (len(summary_models) - 1)
    summary_models = summary_models.reset_index().set_index(['step', 'model'])

    res = create_results(
        RetriesResults,
        input_model_entry,
        input_model_entry,
        res_models,
        rank_type,
        cutoff,
        strictness=strictness,
        summary_models=summary_models,
    )

    res = replace(
        res,
        summary_tool=_modify_summary_tool(res.summary_tool, retry_runs),
    )

    return res


def _modify_summary_tool(summary_tool, retry_runs):
    summary_tool = summary_tool.reset_index()
    number_of_retries_dict = {r.modelentry.model.name: r.number_of_retries for r in retry_runs}
    summary_tool['Number_of_retries'] = summary_tool['model'].map(number_of_retries_dict)

    column_to_move = summary_tool.pop('Number_of_retries')
    summary_tool.insert(1, 'Number_of_retries', column_to_move)

    return summary_tool.set_index(['model'])


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(model, results, number_of_candidates, fraction, strictness, scale, prefix_name):
    if not isinstance(model, Model):
        raise ValueError(
            f'Invalid `model` type: got `{type(model)}`, must be one of pharmpy Model object.'
        )

    if not isinstance(results, ModelfitResults):
        raise ValueError(
            f'Invalid `results` type: got `{type(results)}`, must be one of pharmpy ModelfitResults object.'
        )

    if not isinstance(number_of_candidates, int):
        raise ValueError(
            f'Invalid `number_of_candidates` type: got `{type(number_of_candidates)}`, must be an integer.'
        )
    elif number_of_candidates <= 0:
        raise ValueError(
            f'`number_of_candidates` need to be a positiv integer, not `{number_of_candidates}`'
        )

    if not isinstance(fraction, float):
        raise ValueError(
            f'Invalid `fraction` type: got `{type(fraction)}`, must be an number (float).'
        )

    # STRICTNESS?

    if scale not in SCALES:
        raise ValueError(f'Invalid `scale`: got `{scale}`, must be one of {sorted(SCALES)}.')


@dataclass(frozen=True)
class RetriesResults(ToolResults):
    pass



import pytest

from pharmpy.tools.estmethod.algorithms import _create_base_model, _create_candidate_model
from pharmpy.tools.estmethod.tool import SOLVERS, create_workflow, validate_input
from pharmpy.workflows import ModelEntry


@pytest.mark.parametrize(
    'algorithm, methods, solvers, parameter_uncertainty_methods, no_of_models',
    [
        ('exhaustive', ['foce'], None, None, 1),
        ('exhaustive', ['foce', 'laplace'], None, None, 2),
        ('exhaustive', ['foce', 'imp'], ['lsoda'], None, 2),
        ('exhaustive', ['foce'], 'all', None, len(SOLVERS)),
        ('exhaustive_with_update', ['foce'], None, None, 2),
        ('exhaustive_with_update', ['foce'], None, 'all', 6),
        ('exhaustive_with_update', ['foce', 'laplace'], None, None, 4),
        ('exhaustive_with_update', ['laplace'], None, None, 3),
        ('exhaustive_with_update', ['foce'], ['lsoda'], None, 3),
        ('exhaustive_with_update', ['foce'], 'all', None, len(SOLVERS) * 2 + 1),
        ('exhaustive', ['foce'], None, ['sandwich', 'cpg'], 2),
        ('exhaustive', ['foce', 'imp'], None, ['sandwich', 'cpg'], 4),
    ],
)
def test_algorithm(algorithm, methods, solvers, parameter_uncertainty_methods, no_of_models):
    wf = create_workflow(
        algorithm,
        methods=methods,
        solvers=solvers,
        parameter_uncertainty_methods=parameter_uncertainty_methods,
    )
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models


@pytest.mark.parametrize(
    'est_rec, eval_rec',
    [
        (
            '$ESTIMATION METHOD=COND INTER MAXEVAL=9999 AUTO=1 PRINT=10',
            '$ESTIMATION METHOD=IMP INTER EONLY=1 MAXEVAL=9999 ISAMPLE=10000 ' 'NITER=10 PRINT=10',
        ),
    ],
)
def test_create_base_model(
    load_model_for_test,
    pheno_path,
    est_rec,
    eval_rec,
    parameter_uncertainty_method=None,
    add_eval_after_est=True,
):
    model = load_model_for_test(pheno_path)
    assert len(model.estimation_steps) == 1
    model_entry = ModelEntry.create(model)
    base_model_entry = _create_base_model(
        parameter_uncertainty_method,
        add_eval_after_est,
        model_entry=model_entry,
    )
    base_model = base_model_entry.model
    assert len(base_model.estimation_steps) == 2
    assert base_model.model_code.split('\n')[-5] == est_rec
    assert base_model.model_code.split('\n')[-4] == eval_rec


@pytest.mark.parametrize(
    'method, est_rec, eval_rec',
    [
        (
            'FO',
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9999 AUTO=1 PRINT=10',
            '$ESTIMATION METHOD=IMP INTER EONLY=1 MAXEVAL=9999 ISAMPLE=10000 NITER=10 PRINT=10',
        ),
        (
            'LAPLACE',
            '$ESTIMATION METHOD=COND LAPLACE INTER MAXEVAL=9999 AUTO=1 PRINT=10',
            '$ESTIMATION METHOD=IMP LAPLACE INTER EONLY=1 MAXEVAL=9999 ISAMPLE=10000 '
            'NITER=10 PRINT=10',
        ),
    ],
)
def test_create_candidate_model(
    load_model_for_test,
    pheno_path,
    method,
    est_rec,
    eval_rec,
    parameter_uncertainty_method=None,
    add_eval_after_est=True,
):
    model = load_model_for_test(pheno_path)
    assert len(model.estimation_steps) == 1
    model_entry = ModelEntry.create(model)
    candidate_model_entry = _create_candidate_model(
        '',
        method,
        None,
        parameter_uncertainty_method,
        add_eval_after_est,
        update_inits=False,
        only_evaluation=False,
        model_entry=model_entry,
    )
    candidate_model = candidate_model_entry.model
    assert len(candidate_model.estimation_steps) == 2
    assert candidate_model.model_code.split('\n')[-5] == est_rec
    assert candidate_model.model_code.split('\n')[-4] == eval_rec


@pytest.mark.parametrize(
    (
        'args',
        'exception',
        'match',
    ),
    [
        (
            dict(algorithm='x', methods='all'),
            ValueError,
            'Invalid `algorithm`',
        ),
        (
            dict(algorithm='exhaustive', methods=None, solvers=None),
            ValueError,
            'Invalid search space options',
        ),
        (
            dict(algorithm='exhaustive', solvers=['LSODA']),
            ValueError,
            'Invalid input `model`',
        ),
    ],
)
def test_validate_input(load_model_for_test, pheno_path, args, exception, match):
    model = load_model_for_test(pheno_path)
    kwargs = {**args, 'model': model}
    with pytest.raises(exception, match=match):
        validate_input(**kwargs)



from __future__ import annotations

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.tools.common import ToolResults


@dataclass(frozen=True)
class RUVSearchResults(ToolResults):
    """RUVSearch results class"""

    cwres_models: Optional[Any] = None


def calculate_results(model_entries):
    names = [model_entry.model.name for model_entry in model_entries]
    iterations = [int(name.split('_')[-1]) for name in names]
    iter_dfs = []
    for iteration in range(min(iterations), max(iterations) + 1):
        base_index = names.index(f'base_{iteration}')
        base_model_entry = model_entries[base_index]
        if (
            base_model_entry.modelfit_results is not None
            and base_model_entry.modelfit_results.ofv is not None
        ):
            base_ofv = base_model_entry.modelfit_results.ofv
        else:
            warnings.warn(f'base CWRES model of iteration {iteration} failed.')
            res = RUVSearchResults()
            return res

        iteration_model_entries = [
            model_entry
            for model_entry in model_entries
            if model_entry.model.name.endswith(f'_{iteration}')
            and not model_entry.model.name.startswith('best_ruvsearch')
            and not model_entry.model.name.startswith('base')
        ]

        model_name = []
        model_dofv = []
        model_params = []
        for model_entry in iteration_model_entries:
            name = model_entry.model.name
            model_res = model_entry.modelfit_results
            if model_res is not None and model_res.ofv is not None:
                dofv = base_ofv - model_res.ofv
                if name.startswith('IIV_on_RUV'):
                    param = {'omega': round(model_res.parameter_estimates["IIV_RUV1"], 6)}
                elif name.startswith('power'):
                    param = {'theta': round(model_res.parameter_estimates["power1"], 6)}
                elif name.startswith('time_varying'):
                    param = {'theta': round(model_res.parameter_estimates["time_varying"], 6)}
                else:
                    param = {
                        'sigma_add': round(model_res.parameter_estimates["sigma_add"], 6),
                        'sigma_prop': round(model_res.parameter_estimates["sigma_prop"], 6),
                    }
                a = name.split('_')
                name = '_'.join(a[0:-1])

                model_name.append(name)
                model_dofv.append(dofv)
                model_params.append(param)
            else:
                warnings.warn(
                    f"{name} model has no ofv and will be skipped in {iteration} iteration."
                )
        df = pd.DataFrame(
            {
                'model': model_name,
                'dvid': 1,
                'iteration': iteration,
                'dofv': model_dofv,
                'parameters': model_params,
            }
        )
        if not df.empty:
            iter_dfs.append(df)

    df_final = pd.concat(iter_dfs)
    df_final.set_index(['model', 'dvid', 'iteration'], inplace=True)
    df_final.sort_index(inplace=True)

    return RUVSearchResults(cwres_models=df_final)


def psn_resmod_results(path):
    path = Path(path)
    respath = path / 'resmod_results.csv'
    if not respath.is_file():
        return RUVSearchResults()
    df = pd.read_csv(respath, names=range(40), skiprows=[0], engine='python')
    df[0].fillna(1, inplace=True)
    df[1].fillna(1, inplace=True)
    df.dropna(how='all', axis=1, inplace=True)
    df2 = df[[0, 1, 2, 3]].copy()
    df2 = df2.astype({0: int})
    df2.columns = ['iteration', 'DVID', 'model', 'dOFV']
    df2.set_index(['iteration', 'DVID', 'model'], inplace=True)
    parameters = pd.Series(name='parameters', index=df.index, dtype=object)
    for rowind, row in df.iterrows():
        d = {}
        for i in range(4, len(row)):
            if row[i] is not None and row[i] is not np.nan:
                a = row[i].split('=')
                d[a[0]] = float(a[1])
        parameters[rowind] = d
    parameters.index = df2.index
    df2['parameters'] = parameters
    return RUVSearchResults(cwres_models=df2)



from io import StringIO

import pandas as pd
from numpy import isnan, nan
from pytest import approx

import pharmpy.tools.scm.results as scm


def test_psn_scm_options(testdata):
    options = scm.psn_scm_options(testdata / 'nonmem' / 'scm')
    assert options['included_relations'] is None
    assert options['test_relations'] is None
    assert options['logfile'] == 'scmlog.txt'

    options = scm.psn_scm_options(testdata / 'nonmem' / 'scm' / 'onlyforward_dir1')
    assert options['included_relations'] == {'CL': {'APGR': '2'}, 'V': {'CVD1': '2'}}
    assert options['test_relations'] == {
        'CL': ['WGT', 'APGR', 'CV1', 'CV2', 'CV3'],
        'V': ['CVD1', 'WGT'],
    }
    assert options['logfile'] == 'scmlog.txt'

    options = scm.psn_scm_options(testdata / 'nonmem' / 'scm' / 'backward_dir1')
    assert options['included_relations'] == {
        'CL': {'WGT': '2', 'CV1': '2'},
        'V': {'CV2': '2', 'WGT': '2', 'CVD1': '2'},
    }
    assert options['test_relations'] == {
        'CL': ['WGT', 'APGR', 'CV1', 'CV2', 'CV3'],
        'V': ['CVD1', 'CV2', 'WGT'],
    }
    assert options['logfile'] == 'scmlog.txt'


def test_main(testdata):
    basedir = testdata / 'nonmem' / 'scm'
    dirs = {
        'mer1': basedir / 'mergeofv_dir1',
        'mer2': basedir / 'mergeofv_dir2',
        'onlyf': basedir / 'onlyforward_dir1',
    }

    res1 = scm.psn_scm_results(dirs['mer1'])
    res2 = scm.psn_scm_results(dirs['mer2'])

    assert [float(res1.steps['extended_ofv'].iloc[5])] == approx([219208.47277])
    assert res1.steps['directory'].iloc[0] == 'm1'
    assert isnan(res1.steps['extended_ofv'].iloc[6])
    assert res1.ofv_summary is not None
    assert res1.candidate_summary is not None

    assert [float(res2.steps['ofv_drop'].iloc[2])] == approx([18.5591])
    assert res2.ofv_summary is not None
    assert res2.candidate_summary is not None

    parcovdict = {
        'CLCV2': ('CL', 'CV2'),
        'CLWGT': ('CL', 'WGT'),
        'CLCV3': ('CL', 'CV3'),
        'CLCV1': ('CL', 'CV1'),
        'CLAPGR': ('CL', 'APGR'),
        'VWGT': ('V', 'WGT'),
        'VCVD1': ('V', 'CVD1'),
        'VCV2': ('V', 'CV2'),
    }

    res3 = scm.psn_scm_results(dirs['onlyf'])
    assert res3.steps is not None
    sum1 = res3.ofv_summary
    assert sum1 is not None
    assert [float(x) for x in sum1['pvalue'].values] == approx([nan, nan, 2.22e-13], nan_ok=True)
    assert res3.ofv_summary is not None

    cand1 = res3.candidate_summary
    all_columns = [
        'parameter',
        'covariate',
        'extended_state',
        'N_test',
        'N_ok',
        'N_localmin',
        'N_failed',
        'StepIncluded',
    ]
    correct = """CL,APGR,3, 2,2,   0,0,
CL,CV1,2, 2,2,   0,0,
CL,CV2,2, 2,2,   0,0,
CL,CV3,2, 2,2,   0,0,
CL,WGT,2, 1,1,   0,0,1
CL,WGT,3,  1,1,   0, 0,
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0, 1, 2], names=all_columns, header=None)
    pd.testing.assert_frame_equal(cand1, correct)

    df1 = scm.psn_scm_parse_logfile(
        basedir / 'scmplus_dir1' / 'scmlog.txt',
        {'directory': '/home/kajsa/sandbox/asrscm/asrscm_dir56', 'included_relations': None},
        parcovdict,
    )
    assert df1 is not None
    direct = df1['directory'].iloc[0]
    assert direct == 'rundir/m1' or direct == 'rundir\\m1'
    sum1 = scm.ofv_summary_dataframe(df1)
    assert sum1 is not None
    assert [float(x) for x in sum1['pvalue'].values] == approx(
        [2.26e-10, 0.000002, 0.002178, 0.014192, 0.490010, 9999, 0.027672, 8.15e-24]
    )

    cand1 = scm.candidate_summary_dataframe(df1)
    all_columns = [
        'parameter',
        'covariate',
        'extended_state',
        'N_test',
        'N_ok',
        'N_localmin',
        'N_failed',
        'StepIncluded',
        'StepStashed',
        'StepReadded',
        'BackstepRemoved',
    ]
    correct = """CL,APGR,2,4,2,   2, 0,, 1,3,
CL,CV1,2, 2,2,   0, 0,3,1,3,7
CL,WGT,2, 2,2,   0, 0,2,,, 8
V,CV2,2,  3,3,   0, 0,4,1,3,6
V,CVD1,2, 4,3,   1, 0,,1,3,
V,WGT,2,  1,1,   0, 0,1,,,
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0, 1, 2], names=all_columns, header=None)
    pd.testing.assert_frame_equal(cand1, correct)

    df2 = scm.psn_scm_parse_logfile(
        basedir / 'scm_dir1' / 'scmlog.txt',
        {'directory': '/home/kajsa/sandbox/asrscm/scm_dir7', 'included_relations': None},
        parcovdict,
    )
    assert df2 is not None
    sum2 = scm.ofv_summary_dataframe(df2)
    assert sum2 is not None
    assert [float(x) for x in sum2['ofv_drop'].iloc[2:7].values] == approx(
        [2.90120, 3.00434, 1.18187, 1.18187, 3.00434]
    )
    cand2 = scm.candidate_summary_dataframe(df2)
    assert cand2 is not None
    assert [int(t) for t in cand2['N_test']] == [6, 3, 2, 4, 5, 1]

    df3 = scm.psn_scm_parse_logfile(
        basedir / 'backward_dir1' / 'scmlog.txt',
        {
            'directory': '/home/kajsa/sandbox/asrscm/scm_dir10',
            'included_relations': {
                'CL': {'WGT': '2', 'CV1': '2'},
                'V': {'CV2': '2', 'WGT': '2', 'CVD1': '2'},
            },
        },
        parcovdict,
    )
    assert df3 is not None
    assert scm.ofv_summary_dataframe(df3) is not None
    cand1 = scm.candidate_summary_dataframe(df3)

    correct = """V,CVD1,2,1
V,CV2,2,2
CL,CV1,2,3
"""
    correct = pd.read_csv(
        StringIO(correct),
        index_col=[0, 1, 2],
        names=['parameter', 'covariate', 'extended_state', 'BackstepRemoved'],
        header=None,
    )
    pd.testing.assert_frame_equal(cand1, correct)

    df4 = scm.psn_scm_parse_logfile(
        basedir / 'localmin.logf',
        {'directory': '/home/kajsa/sandbox/asrscm/asrscm_dir5', 'included_relations': None},
        parcovdict,
    )
    assert df4 is not None
    assert scm.ofv_summary_dataframe(df4) is not None
    assert scm.candidate_summary_dataframe(df4) is not None

    gof1 = scm.psn_scm_parse_logfile(
        basedir / 'gofofv_dir1' / 'scmlog.txt',
        {'directory': '/home/kajsa/sandbox/asrscm/scm_dir8', 'included_relations': None},
        parcovdict,
    )
    assert gof1 is not None
    assert scm.ofv_summary_dataframe(gof1) is not None
    assert scm.candidate_summary_dataframe(gof1) is not None


def test_parse_runtable_block_gof_pval(testdata):
    basedir = testdata / 'nonmem' / 'scm'
    step = {
        'b_pval': basedir / 'log_steps' / 'backward_pval_1.txt',
        'f_pval': basedir / 'log_steps' / 'forward_pval_1.txt',
        'm_pos': basedir / 'log_steps' / 'forward_pval_2.txt',
        'm_neg': basedir / 'log_steps' / 'forward_pval_3.txt',
        'f_nosignif': basedir / 'log_steps' / 'forward_pval_4.txt',
        'b_nosignif': basedir / 'log_steps' / 'backward_pval_2.txt',
    }

    parcov = {
        'CLCV2': ('CL', 'CV2'),
        'CLWGT': ('CL', 'WGT'),
        'CLCV3': ('CL', 'CV3'),
        'CLCV1': ('CL', 'CV1'),
        'CLAPGR': ('CL', 'APGR'),
        'VWGT': ('V', 'WGT'),
        'VCVD1': ('V', 'CVD1'),
        'VCV2': ('V', 'CV2'),
    }

    columns = [
        'model',
        'parameter',
        'covariate',
        'extended_state',
        'reduced_ofv',
        'extended_ofv',
        'ofv_drop',
        'delta_df',
        'pvalue',
        'is_backward',
        'extended_significant',
    ]

    with open(step['b_pval']) as tab:
        block = [row for row in tab]

    correct = """0,CLWGT-1,CL,WGT,,657.31367,653.74502,3.56865,1,0.058880,True,False
1,VCV2-1,V,CV2,,654.03358,653.74502,0.28856,1,0.59115,True,False
2,VWGT-1,V,WGT,,729.06993,653.74502,75.32491,1,3.99e-18,True,True
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0], names=columns, header=None)
    table = scm.parse_runtable_block(block, parcov_dictionary=parcov)
    pd.testing.assert_frame_equal(table, correct)

    with open(step['f_pval']) as tab:
        block = [row for row in tab]

    correct = """0,CLAPGR-2,CL,APGR,2,724.0,725.6143599999999,-1.61436,9,9999.0,False,False
1,CLCV1-2,CL,CV1,2,724.0,723.11595,0.88405,1,0.34709,False,False
2,CLWGT-2,CL,WGT,2,724.0,699.47082,24.52918,1,1e-06,False,True
3,VCV2-2,V,CV2,2,724.0,724.9772,-0.9772,1,9999.0,False,False
4,VCVD1-2,V,CVD1,2,724.0,723.0435,0.9565,1,0.32807,False,False
5,VWGT-2,V,WGT,2,724.0,685.3777,38.6223,1,5.14e-10,False,True
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0], names=columns, header=None)
    table = scm.parse_runtable_block(block, parcov_dictionary=parcov)
    table = table.astype({'extended_state': 'int64'})
    pd.testing.assert_frame_equal(table, correct)

    # log files with merged base and new ofv. Small test to ensure correct
    # unmerging of column values
    with open(step['m_pos']) as tab:
        block = [row for row in tab]
    table = scm.parse_runtable_block(block)
    assert [float(x) for x in table['extended_ofv'].iloc[5:8].values] == approx(
        [219208.47277, nan, 217228.28027], nan_ok=True
    )

    with open(step['m_neg']) as tab:
        block = [row for row in tab]
    table = scm.parse_runtable_block(block)
    assert [float(x) for x in table['extended_ofv'].iloc[0:3].values] == approx(
        [-10620.89249, -10638.97128, -10629.37133]
    )

    with open(step['f_nosignif']) as tab:
        block = [row for row in tab]
    table = scm.parse_runtable_block(block)
    assert [float(x) for x in table['pvalue'].iloc[0:2].values] == approx([9999, 9999])
    assert [x for x in table['extended_significant'].iloc[0:2].values] == [False, False]

    with open(step['b_nosignif']) as tab:
        block = [row for row in tab]
    table = scm.parse_runtable_block(block)
    assert [float(table['pvalue'].iloc[0])] == approx([8.15e-24])
    assert table['extended_significant'].iloc[0]  # assert is True


def test_parse_runtable_block_gof_ofv(testdata):
    basedir = testdata / 'nonmem' / 'scm'
    step = {
        'b_ofv': basedir / 'log_steps' / 'backward_ofv_1.txt',
        'f_ofv': basedir / 'log_steps' / 'forward_ofv_1.txt',
    }

    parcov = {
        'CLCV2': ('CL', 'CV2'),
        'CLWGT': ('CL', 'WGT'),
        'CLCV3': ('CL', 'CV3'),
        'CLCV1': ('CL', 'CV1'),
        'CLAPGR': ('CL', 'APGR'),
        'VWGT': ('V', 'WGT'),
        'VCVD1': ('V', 'CVD1'),
        'VCV2': ('V', 'CV2'),
    }

    columns = [
        'model',
        'parameter',
        'covariate',
        'extended_state',
        'reduced_ofv',
        'extended_ofv',
        'ofv_drop',
        'goal_ofv_drop',
        'is_backward',
        'extended_significant',
    ]

    with open(step['b_ofv']) as tab:
        block = [row for row in tab]

    correct = """0,CLCV1-1,CL,CV1,,587.20249,584.15462,3.04787,7.63,True,False
1,CLWGT-1,CL,WGT,,622.67113,584.15462,38.51651,7.63,True,True
2,VCV2-1,V,CV2,,587.15896,584.15462,3.00434,7.63,True,False
3,VWGT-1,V,WGT,,669.56839,584.15462,85.41377,7.63,True,True
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0], names=columns, header=None)
    table = scm.parse_runtable_block(block, parcov_dictionary=parcov)
    pd.testing.assert_frame_equal(table, correct)

    with open(step['f_ofv']) as tab:
        block = [row for row in tab]

    correct = """0,CLAPGR-2,CL,APGR,2,725.602,705.46995,20.13205,16.92,False,True
1,CLCV1-2,CL,CV1,2,725.602,722.68945,2.91255,2.84,False,True
2,CLWGT-2,CL,WGT,2,725.602,672.98298,52.61901999999999,2.84,False,True
3,VCV2-2,V,CV2,2,725.602,724.65062,0.9513799999999999,2.84,False,False
4,VCVD1-2,V,CVD1,2,725.602,717.75649,7.845510000000001,2.84,False,True
5,VWGT-2,V,WGT,2,725.602,629.27205,96.32995,2.84,False,True
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0], names=columns, header=None)
    table = scm.parse_runtable_block(block, parcov_dictionary=parcov)
    table = table.astype({'extended_state': 'int64'})
    pd.testing.assert_frame_equal(table, correct)


def test_parse_mixed_block():
    block = [
        'Using user-defined ofv change criteria',
        'Degree of freedom  |  Required ofv change',
        '         1         -          7.63',
        '         2         -          10.21',
        '         3         -          11.34',
        '         4         -          13.28',
        '         5         -          15.09',
        '         6         -          16.81',
        '         7         -          18.48',
        '         8         -          20.09',
        '         9         -          21.67',
        '         10         -          23.21',
        r'Model directory' + ' /sandbox/asrscm/scm_dir8/backward_scm_dir1/scm_dir1/scm_dir1/m1',
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 == '/sandbox/asrscm/scm_dir8/backward_scm_dir1/scm_dir1/scm_dir1/m1'
    assert readded == []
    assert stashed == []
    assert included_relations == {}

    block = [
        'Forward search done. Starting backward search inside forward top level directory',
        r'Model directory /home/run1080scm_3/backward_scm_dir1/m1',
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 == '/home/run1080scm_3/backward_scm_dir1/m1'
    assert readded == []
    assert stashed == []
    assert included_relations == {}

    block = [
        'Taking a step forward in adaptive scope reduction scm after reducing scope '
        + r'with 15 relations : '
        + r'CL-SEX,CL-QUAT,KM-AGE,KM-RACE,KM-SEX,KM-QUAT,V1-AGE,V1-DOSE,'
        + r'V1-QUAT,V2-AGE,V2-RACE,VM-AGE,VM-RACE,VM-SEX,VM-QUAT',
        'Model directory /home/Drug/run15asr_2.dir/rundir/reduced_forward_scm_dir2/m1',
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 == '/home/Drug/run15asr_2.dir/rundir/reduced_forward_scm_dir2/m1'
    assert readded == []
    assert stashed == [
        ('CL', 'SEX'),
        ('CL', 'QUAT'),
        ('KM', 'AGE'),
        ('KM', 'RACE'),
        ('KM', 'SEX'),
        ('KM', 'QUAT'),
        ('V1', 'AGE'),
        ('V1', 'DOSE'),
        ('V1', 'QUAT'),
        ('V2', 'AGE'),
        ('V2', 'RACE'),
        ('VM', 'AGE'),
        ('VM', 'RACE'),
        ('VM', 'SEX'),
        ('VM', 'QUAT'),
    ]
    assert included_relations == {}

    block = [
        'Taking a step forward in adaptive scope reduction '
        + r'scm after reducing scope with 1 relations : V2-SEX',
        'Model directory /home/shared/rundir/reduced_forward_scm_dir3/m1',
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 == '/home/shared/rundir/reduced_forward_scm_dir3/m1'
    assert readded == []
    assert stashed == [('V2', 'SEX')]
    assert included_relations == {}

    block = [
        'Taking a step forward in adaptive scope reduction '
        + r'scm after reducing scope with 1 relations : V1-RACE'
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 is None
    assert readded == []
    assert stashed == [('V1', 'RACE')]
    assert included_relations == {}

    block = ['No models to test, there are no relations to add.']
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 is None
    assert readded == []
    assert stashed == []
    assert included_relations == {}

    block = [
        'Scope reduction requested in adaptive scope reduction scm after forward step 6 but',
        'no relation was chosen for inclusion by scm in this step:',
        'adaptive scope reduction scm forward search with reduced scope is done.',
        'Included relations so far:  CL-CRCL-5,CL-DOSE-2,CL-RACE-2,V1-SEX-2,V2-QUAT-2',
        'Re-testing 19 relations after adaptive scope reduction '
        + r'scm reduced forward search : '
        + r'CL-SEX,CL-QUAT,KM-AGE,KM-RACE,KM-SEX,KM-QUAT,V1-AGE,V1-DOSE,'
        + r'V1-QUAT,V2-AGE,V2-RACE,VM-AGE,VM-RACE,VM-SEX,VM-QUAT,V2-SEX,CL-AGE,V2-DOSE,V1-RACE',
        'Model directory /home/run15asr_2.dir/rundir/readded_forward_scm_dir7/m1',
    ]
    m1, readded, stashed, included_relations = scm.parse_mixed_block(block)
    assert m1 == '/home/run15asr_2.dir/rundir/readded_forward_scm_dir7/m1'
    assert readded == [
        ('CL', 'SEX'),
        ('CL', 'QUAT'),
        ('KM', 'AGE'),
        ('KM', 'RACE'),
        ('KM', 'SEX'),
        ('KM', 'QUAT'),
        ('V1', 'AGE'),
        ('V1', 'DOSE'),
        ('V1', 'QUAT'),
        ('V2', 'AGE'),
        ('V2', 'RACE'),
        ('VM', 'AGE'),
        ('VM', 'RACE'),
        ('VM', 'SEX'),
        ('VM', 'QUAT'),
        ('V2', 'SEX'),
        ('CL', 'AGE'),
        ('V2', 'DOSE'),
        ('V1', 'RACE'),
    ]
    assert stashed == []
    assert included_relations == {
        'CL': {'CRCL': '5', 'DOSE': '2', 'RACE': '2'},
        'V1': {'SEX': '2'},
        'V2': {'QUAT': '2'},
    }


def test_parse_chosen_relation_block():
    block = [
        'Parameter-covariate relation chosen in this forward step: PHA-DISDUR-5 \n',
        'CRITERION              PVAL < 0.01 \n',
        'BASE_MODEL_OFV      215748.07194\n',
        'CHOSEN_MODEL_OFV    215570.57352\n',
        'Relations included after this step:\n',
        'PHA     DISDUR-5MENS-2 \n',
        'PHIA\n',
        'PHIB    MENS-2\n',
        'SLP\n',
    ]
    chosen, criterion, included_relations = scm.parse_chosen_relation_block(block)
    assert chosen == {'parameter': 'PHA', 'covariate': 'DISDUR', 'state': '5'}
    assert criterion == {'gof': 'PVAL', 'pvalue': '0.01', 'is_backward': False}
    assert included_relations == {
        'PHA': {'DISDUR': '5', 'MENS': '2'},
        'PHIA': {},
        'PHIB': {'MENS': '2'},
        'SLP': {},
    }
    block = [
        'Parameter-covariate relation chosen in this forward step: PHA-WT-5',
        'CRITERION              PVAL < 0.01',
        'BASE_MODEL_OFV      215321.70954',
        'CHOSEN_MODEL_OFV    215306.67698',
        'Relations included after this step:',
        'PHA     AGE-5   BMI-5   DISDUR-5MENS-2  WT-5  ',
        'PHIA    MENS-2  ',
        'PHIB    GEOREG-2MENS-2  ',
        'SLP     GEOREG-2',
    ]

    chosen, criterion, included_relations = scm.parse_chosen_relation_block(block)
    assert chosen == {'parameter': 'PHA', 'covariate': 'WT', 'state': '5'}
    assert criterion == {'gof': 'PVAL', 'pvalue': '0.01', 'is_backward': False}

    assert included_relations == {
        'PHA': {'AGE': '5', 'BMI': '5', 'DISDUR': '5', 'MENS': '2', 'WT': '5'},
        'PHIA': {'MENS': '2'},
        'PHIB': {'GEOREG': '2', 'MENS': '2'},
        'SLP': {'GEOREG': '2'},
    }

    block = [
        'Parameter-covariate relation chosen in this forward step: --',
        'CRITERION              PVAL < 0.01',
    ]

    chosen, criterion, included_relations = scm.parse_chosen_relation_block(block)
    assert chosen == {}
    assert criterion == {'gof': 'PVAL', 'pvalue': '0.01', 'is_backward': False}

    assert included_relations == {}

    block = [
        'Parameter-covariate relation chosen in this backward step: --',
        'CRITERION              PVAL > 0.001',
    ]
    chosen, criterion, included_relations = scm.parse_chosen_relation_block(block)
    assert chosen == {}
    assert criterion == {'gof': 'PVAL', 'pvalue': '0.001', 'is_backward': True}

    assert included_relations == {}

    block = [
        'Parameter-covariate relation chosen in this forward step: V-CV2-2',
        'CRITERION              OFV',
        'BASE_MODEL_OFV         587.15896',
        'CHOSEN_MODEL_OFV       584.15462',
        'Relations included after this step:',
        'CL      CV1-2   WGT-2   ',
        'V       CV2-2   WGT-2',
    ]
    chosen, criterion, included_relations = scm.parse_chosen_relation_block(block)
    assert chosen == {'parameter': 'V', 'covariate': 'CV2', 'state': '2'}
    assert criterion == {'gof': 'OFV', 'pvalue': None, 'is_backward': False}

    assert included_relations == {'CL': {'CV1': '2', 'WGT': '2'}, 'V': {'CV2': '2', 'WGT': '2'}}



from pharmpy.internals.fs.cwd import chdir
from pharmpy.tools import read_results, run_modelsearch


def test_serialization(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_modelsearch('ABSORPTION(ZO)', 'exhaustive', model=start_modelres[0])

        rundir = tmp_path / 'modelsearch_dir1'

        res_serialized = read_results(rundir / 'results.json')

        # This is a temporary workaround since floats are serialized with 15 precision, should be 17 for correct
        # round-trip
        cols_float = res.summary_tool.select_dtypes(include=['float64']).columns
        cols_round = {col: 15 for col in cols_float}
        summary_tool_rounded = res.summary_tool.round(cols_round)
        summary_tool_serialized_rounded = res_serialized.summary_tool.round(cols_round)

        assert summary_tool_rounded.equals(summary_tool_serialized_rounded)
        assert res.tool_database.path == res_serialized.tool_database.path
        assert (
            res.tool_database.model_database.path
            == res_serialized.tool_database.model_database.path
        )



from pharmpy.internals.module.lazy import LazyImport


def test_dir():
    import pharmpy.internals.module.lazy

    module = LazyImport('x', {}, 'pharmpy.internals.module.lazy')
    assert dir(pharmpy.internals.module.lazy) == dir(module)


def test_getattr():
    module = LazyImport('x', {}, 'pharmpy.internals.module.lazy')
    assert getattr(module, 'LazyImport') is LazyImport


def test_submodule():
    import os

    module = LazyImport('x', {}, 'os', attr='path')
    assert dir(module) == dir(os.path)



r"""
Data
====

.. list-table:: Options for the data module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``na_values``
     - ``[-99]``
     - List
     - Data values to be converted to NA when reading in data
   * - ``na_rep``
     - ``'-99'``
     - str
     - Data value to convert NA to when writing data

"""

import pharmpy.config as config


class DataConfiguration(config.Configuration):
    module = 'pharmpy.data'
    na_values = config.ConfigItem(
        [-99], 'List of data values to be converted to NA when reading data'
    )
    na_rep = config.ConfigItem('-99', 'What to replace NA with in written datasets')


conf = DataConfiguration()


class DatasetError(Exception):
    """Exception for errors in the dataset"""

    pass


class DatasetWarning(Warning):
    """Warning for recoverable issues in the dataset"""

    pass



import pharmpy.workflows.log


def test_log():
    client = pharmpy.workflows.log.Log()
    client = client.log_error("help!")
    client = client.log_warning("an annoying warning")
    df = client.to_dataframe()
    assert list(df['category']) == ['ERROR', 'WARNING']
    assert list(df['message']) == ['help!', 'an annoying warning']



"""
=======
Pharmpy
=======

Pharmpy is a python package for pharmacometric modeling.

Definitions
===========
"""

__version__ = '0.106.0'

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



from __future__ import annotations

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def parse(expr: Union[int, float, str, sympy.Expr]) -> sympy.Expr:
    ns = {'Q': sympy.Symbol('Q'), 'LT': sympy.Symbol('LT'), 'N': sympy.Symbol('N')}
    expr = sympy.sympify(expr, locals=ns)
    assert isinstance(expr, sympy.Expr)
    return expr



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Mapping, Optional, Union, overload

from pharmpy.internals.immutable import Immutable, frozenmapping

if TYPE_CHECKING:
    import pandas as pd
else:
    from pharmpy.deps import pandas as pd


SUPPORTED_SOLVERS = frozenset(('CVODES', 'DGEAR', 'DVERK', 'IDA', 'LSODA', 'LSODI'))


class Step(Immutable):
    def __init__(
        self,
        solver: Optional[str] = None,
        solver_rtol: Optional[int] = None,
        solver_atol: Optional[int] = None,
        tool_options: Optional[frozenmapping[str, Any]] = None,
    ):
        self._solver = solver
        self._solver_rtol = solver_rtol
        self._solver_atol = solver_atol
        self._tool_options = tool_options

    @staticmethod
    def _canonicalize_solver(solver):
        if solver is not None:
            solver = solver.upper()
        if not (solver is None or solver in SUPPORTED_SOLVERS):
            raise ValueError(
                f"Unknown solver {solver}. Recognized solvers are {sorted(SUPPORTED_SOLVERS)}."
            )
        return solver

    @staticmethod
    def _canonicalize_tool_options(tool_options):
        if tool_options is None:
            tool_options = frozenmapping({})
        else:
            tool_options = frozenmapping(tool_options)
        return tool_options

    @property
    def solver(self) -> Union[str, None]:
        """Numerical solver to use when numerically solving the ODE system
        Supported solvers and their corresponding NONMEM ADVAN

        +----------------------------+------------------+
        | Solver                     | NONMEM ADVAN     |
        +============================+==================+
        | CVODES                     | ADVAN14          |
        +----------------------------+------------------+
        | DGEAR                      | ADVAN8           |
        +----------------------------+------------------+
        | DVERK                      | ADVAN6           |
        +----------------------------+------------------+
        | IDA                        | ADVAN15          |
        +----------------------------+------------------+
        | LSODA                      | ADVAN13          |
        +----------------------------+------------------+
        | LSODI                      | ADVAN9           |
        +----------------------------+------------------+
        """
        return self._solver

    @property
    def solver_rtol(self) -> Union[int, None]:
        """Relative tolerance for numerical ODE system solver"""
        return self._solver_rtol

    @property
    def solver_atol(self) -> Union[int, None]:
        """Absolute tolerance for numerical ODE system solver"""
        return self._solver_atol

    @property
    def tool_options(self) -> Union[frozenmapping[str, Any], None]:
        """Dictionary of tool specific options"""
        return self._tool_options

    def _add_to_dict(self, d):
        if self._tool_options is not None:
            tool_options = dict(self._tool_options)
        else:
            tool_options = self._tool_options  # self._tool_options is None
        d['solver'] = self._solver
        d['solver_rtol'] = self._solver_rtol
        d['solver_atol'] = self._solver_atol
        d['tool_options'] = tool_options

    @staticmethod
    def _adjust_dict(d):
        del d['class']
        if isinstance(d['tool_options'], dict):
            d['tool_options'] = frozenmapping(d['tool_options'])

    def _partial_repr(self):
        solver = f"'{self.solver}'" if self.solver is not None else self.solver
        return (
            f"solver={solver}, "
            f"solver_rtol={self.solver_rtol}, solver_atol={self.solver_atol}, "
            f"tool_options={self.tool_options}"
        )

    def __eq__(self, other):
        # NOTE: No need to test class here
        return (
            self.solver == other.solver
            and self.solver_rtol == other.solver_rtol
            and self.solver_atol == other.solver_atol
            and self.tool_options == other.tool_options
        )

    def __hash__(self):
        return hash(
            (
                self._solver,
                self._solver_rtol,
                self._solver_atol,
                self._tool_options,
            )
        )


class EstimationStep(Step):
    """Definition of one estimation operation"""

    """Supported estimation methods
    """
    supported_methods = frozenset(('FO', 'FOCE', 'ITS', 'IMPMAP', 'IMP', 'SAEM', 'BAYES'))
    supported_parameter_uncertainty_methods = frozenset(('SANDWICH', 'CPG', 'OFIM', 'EFIM'))

    def __init__(
        self,
        method: str,
        interaction: bool = False,
        parameter_uncertainty_method: Optional[str] = None,
        evaluation: bool = False,
        maximum_evaluations: Optional[int] = None,
        laplace: bool = False,
        isample: Optional[int] = None,
        niter: Optional[int] = None,
        auto: Optional[bool] = None,
        keep_every_nth_iter: Optional[int] = None,
        residuals: Optional[tuple[str, ...]] = None,
        predictions: Optional[tuple[str, ...]] = None,
        solver: Optional[str] = None,
        solver_rtol: Optional[int] = None,
        solver_atol: Optional[int] = None,
        tool_options: Optional[frozenmapping[str, Any]] = None,
        eta_derivatives: Optional[tuple[str, ...]] = None,
        epsilon_derivatives: Optional[tuple[str, ...]] = None,
    ):
        self._method = method
        self._interaction = interaction
        self._parameter_uncertainty_method = parameter_uncertainty_method
        self._evaluation = evaluation
        self._maximum_evaluations = maximum_evaluations
        self._laplace = laplace
        self._isample = isample
        self._niter = niter
        self._auto = auto
        self._keep_every_nth_iter = keep_every_nth_iter
        self._residuals = residuals
        self._predictions = predictions
        self._eta_derivatives = eta_derivatives
        self._epsilon_derivatives = epsilon_derivatives
        super().__init__(
            solver=solver,
            solver_rtol=solver_rtol,
            solver_atol=solver_atol,
            tool_options=tool_options,
        )

    @classmethod
    def create(
        cls,
        method: str,
        interaction: bool = False,
        parameter_uncertainty_method: Optional[str] = None,
        evaluation: bool = False,
        maximum_evaluations: Optional[int] = None,
        laplace: bool = False,
        isample: Optional[int] = None,
        niter: Optional[int] = None,
        auto: Optional[bool] = None,
        keep_every_nth_iter: Optional[int] = None,
        residuals: Optional[Sequence[str]] = None,
        predictions: Optional[Sequence[str]] = None,
        solver: Optional[str] = None,
        solver_rtol: Optional[int] = None,
        solver_atol: Optional[int] = None,
        tool_options: Optional[Mapping[str, Any]] = None,
        eta_derivatives: Optional[Sequence[str]] = None,
        epsilon_derivatives: Optional[Sequence[str]] = None,
    ):
        method = EstimationStep._canonicalize_and_check_method(method)
        if maximum_evaluations is not None and maximum_evaluations < 1:
            raise ValueError(
                'Number of maximum evaluations must be more than one, use '
                'evaluation=True or tool_options for special cases (e.g. 0 and -1'
                'in NONMEM)'
            )
        if residuals is None:
            residuals = ()
        else:
            residuals = tuple(residuals)
        if predictions is None:
            predictions = ()
        else:
            predictions = tuple(predictions)
        if parameter_uncertainty_method is not None:
            parameter_uncertainty_method = parameter_uncertainty_method.upper()
        if not (
            parameter_uncertainty_method is None
            or parameter_uncertainty_method
            in EstimationStep.supported_parameter_uncertainty_methods
        ):
            raise ValueError(
                f"Unknown parameter uncertainty method {parameter_uncertainty_method}. "
                f"Recognized methods are {sorted(EstimationStep.supported_parameter_uncertainty_methods)}."
            )
        solver = Step._canonicalize_solver(solver)
        tool_options = Step._canonicalize_tool_options(tool_options)
        if eta_derivatives is None:
            eta_derivatives = ()
        else:
            eta_derivatives = tuple(eta_derivatives)
        if epsilon_derivatives is None:
            epsilon_derivatives = ()
        else:
            epsilon_derivatives = tuple(epsilon_derivatives)
        return cls(
            method=method,
            interaction=interaction,
            parameter_uncertainty_method=parameter_uncertainty_method,
            evaluation=evaluation,
            maximum_evaluations=maximum_evaluations,
            laplace=laplace,
            isample=isample,
            niter=niter,
            auto=auto,
            keep_every_nth_iter=keep_every_nth_iter,
            residuals=residuals,
            predictions=predictions,
            solver=solver,
            solver_rtol=solver_rtol,
            solver_atol=solver_atol,
            tool_options=tool_options,
            eta_derivatives=eta_derivatives,
            epsilon_derivatives=epsilon_derivatives,
        )

    def replace(self, **kwargs) -> EstimationStep:
        """Derive a new EstimationStep with new properties"""
        d = {key[1:]: value for key, value in self.__dict__.items()}
        d.update(kwargs)
        new = EstimationStep.create(**d)
        return new

    @staticmethod
    def _canonicalize_and_check_method(method: str) -> str:
        method = method.upper()
        if method not in EstimationStep.supported_methods:
            raise ValueError(
                f'EstimationStep: {method} not recognized. Use any of {sorted(EstimationStep.supported_methods)}.'
            )
        return method

    @property
    def method(self) -> str:
        """Name of the estimation method"""
        return self._method

    @property
    def maximum_evaluations(self) -> Union[int, None]:
        """Maximum allowable number of evaluations of the objective function"""
        return self._maximum_evaluations

    @property
    def interaction(self) -> bool:
        """Preserve eta-epsilon interaction in the computation of the objective function"""
        return self._interaction

    @property
    def evaluation(self) -> bool:
        """Only perform model evaluation"""
        return self._evaluation

    @property
    def parameter_uncertainty_method(self) -> Union[str, None]:
        """Method to use when estimating parameter uncertainty.
        Supported methods and their corresponding NMTRAN code:

        +----------------------------+-----------------------+
        | Method                     | NMTRAN                |
        +============================+=======================+
        | Sandwich                   | $COVARIANCE           |
        +----------------------------+-----------------------+
        | Cross-product gradient     | $COVARIANCE MATRIX=S  |
        +----------------------------+-----------------------+
        | Observed FIM               | $COVARIANCE MATRIX=R  |
        +----------------------------+-----------------------+
        | Expected FIM               | $DESIGN               |
        +----------------------------+-----------------------+

        By default the following options are appended:
        UNCONDITIONAL: The uncertainty step is implemented regardless of how the estimation step terminates.
        PRINT=E: Print the eigenvalues of the correlation matrix.
        PRECOND=1: Perform up to 1 preconditioning cycle on the R matrix.
        """
        return self._parameter_uncertainty_method

    @property
    def laplace(self) -> bool:
        """Use the laplacian method"""
        return self._laplace

    @property
    def isample(self) -> Union[int, None]:
        """Number of samples per subject (or similar) for EM methods"""
        return self._isample

    @property
    def niter(self) -> Union[int, None]:
        """Number of iterations for EM methods"""
        return self._niter

    @property
    def auto(self) -> Union[bool, None]:
        """Let estimation tool automatically add settings"""
        return self._auto

    @property
    def keep_every_nth_iter(self) -> Union[int, None]:
        """Keep results for every nth iteration"""
        return self._keep_every_nth_iter

    @property
    def residuals(self) -> Union[tuple[str, ...], None]:
        """List of residuals to calculate"""
        return self._residuals

    @property
    def predictions(self) -> Union[tuple[str, ...], None]:
        """List of predictions to estimate"""
        return self._predictions

    @property
    def eta_derivatives(self) -> Union[tuple[str, ...], None]:
        """List of names of etas for which to calculate derivatives"""
        return self._eta_derivatives

    @property
    def epsilon_derivatives(self) -> Union[tuple[str, ...], None]:
        """List of names of epsilons for which to calculate derivatives"""
        return self._epsilon_derivatives

    @property
    def tool_options(self) -> Union[frozenmapping[str, Any], None]:
        """Dictionary of tool specific options"""
        return self._tool_options

    def __eq__(self, other):
        return (
            isinstance(other, EstimationStep)
            and self.method == other.method
            and self.interaction == other.interaction
            and self.parameter_uncertainty_method == other.parameter_uncertainty_method
            and self.evaluation == other.evaluation
            and self.maximum_evaluations == other.maximum_evaluations
            and self.laplace == other.laplace
            and self.isample == other.isample
            and self.niter == other.niter
            and self.auto == other.auto
            and self.keep_every_nth_iter == other.keep_every_nth_iter
            and super().__eq__(other)
        )

    def __hash__(self):
        return hash(
            (
                self._method,
                self._interaction,
                self._parameter_uncertainty_method,
                self._evaluation,
                self._maximum_evaluations,
                self._laplace,
                self._isample,
                self._niter,
                self._auto,
                self._keep_every_nth_iter,
                super().__hash__(),
            )
        )

    def to_dict(self) -> dict[str, Any]:
        d = {
            'class': 'EstimationStep',
            'method': self._method,
            'interaction': self._interaction,
            'parameter_uncertainty_method': self._parameter_uncertainty_method,
            'evaluation': self._evaluation,
            'maximum_evaluations': self._maximum_evaluations,
            'laplace': self._laplace,
            'isample': self._isample,
            'niter': self._niter,
            'auto': self._auto,
            'keep_every_nth_iter': self._keep_every_nth_iter,
        }
        super()._add_to_dict(d)
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> EstimationStep:
        d = dict(d)
        Step._adjust_dict(d)
        return cls(**d)

    def __repr__(self):
        parameter_uncertainty_method = (
            f"'{self.parameter_uncertainty_method}'"
            if self.parameter_uncertainty_method is not None
            else self.parameter_uncertainty_method
        )
        return (
            f"EstimationStep('{self.method}', interaction={self.interaction}, "
            f"parameter_uncertainty_method={parameter_uncertainty_method}, evaluation={self.evaluation}, "
            f"maximum_evaluations={self.maximum_evaluations}, laplace={self.laplace}, "
            f"isample={self.isample}, niter={self.niter}, auto={self.auto}, "
            f"keep_every_nth_iter={self.keep_every_nth_iter}, {super()._partial_repr()})"
        )


class SimulationStep(Step):
    """Definition of one simulation operation"""

    def __init__(
        self,
        n: int = 1,
        seed: int = 64206,
        solver: Optional[str] = None,
        solver_rtol: Optional[int] = None,
        solver_atol: Optional[int] = None,
        tool_options: Optional[frozenmapping[str, Any]] = None,
    ):
        self._n = n
        self._seed = seed
        super().__init__(
            solver=solver,
            solver_rtol=solver_rtol,
            solver_atol=solver_atol,
            tool_options=tool_options,
        )

    @classmethod
    def create(
        cls,
        n: int = 1,
        seed: int = 64206,
        solver: Optional[str] = None,
        solver_rtol: Optional[int] = None,
        solver_atol: Optional[int] = None,
        tool_options: Optional[Mapping[str, Any]] = None,
    ):
        if n < 1:
            raise ValueError("Need at least one replicate in SimulationStep")
        return cls(n=n, seed=seed)

    def replace(self, **kwargs) -> SimulationStep:
        """Derive a new SimulationStep with new properties"""
        d = {key[1:]: value for key, value in self.__dict__.items()}
        d.update(kwargs)
        new = SimulationStep.create(**d)
        return new

    @property
    def n(self) -> int:
        """Number of simulation replicates"""
        return self._n

    @property
    def seed(self) -> int:
        """Random seed"""
        return self._seed

    def __eq__(self, other):
        return (
            isinstance(other, SimulationStep)
            and self.n == other.n
            and self.seed == other.seed
            and super().__eq__(other)
        )

    def __hash__(self):
        return hash((self._n, self._seed, super().__hash__()))

    def to_dict(self) -> dict[str, Any]:
        d = {
            'class': 'SimulationStep',
            'n': self._n,
            'seed': self._seed,
        }
        super()._add_to_dict(d)
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> SimulationStep:
        d = dict(d)
        Step._adjust_dict(d)
        return cls(**d)

    def __repr__(self):
        return f"SimulationStep(n={self._n}, seed={self._seed}, {super()._partial_repr()})"


class EstimationSteps(Sequence, Immutable):
    """A sequence of estimation steps

    Parameters
    ----------
    steps : iterable or None
        Used for initialization
    """

    def __init__(self, steps: tuple[Union[EstimationStep, SimulationStep], ...] = ()):
        self._steps = steps

    @classmethod
    def create(cls, steps: Optional[Sequence[Union[EstimationStep, SimulationStep]]] = None):
        if steps is None:
            steps = ()
        else:
            steps = tuple(steps)
        return EstimationSteps(steps=steps)

    def replace(self, **kwargs) -> EstimationSteps:
        steps = kwargs.get('steps', self._steps)
        return EstimationSteps.create(steps)

    @overload
    def __getitem__(self, i: int) -> EstimationStep:
        ...

    @overload
    def __getitem__(self, i: slice) -> EstimationSteps:
        ...

    def __getitem__(
        self, i: Union[int, slice]
    ) -> Union[EstimationStep, SimulationStep, EstimationSteps]:
        if isinstance(i, slice):
            return EstimationSteps(self._steps[i])
        return self._steps[i]

    def __add__(self, other) -> EstimationSteps:
        if isinstance(other, EstimationSteps):
            return EstimationSteps(self._steps + other._steps)
        elif isinstance(other, EstimationStep) or isinstance(other, SimulationStep):
            return EstimationSteps(self._steps + (other,))
        else:
            return EstimationSteps(self._steps + tuple(other))

    def __radd__(self, other) -> EstimationSteps:
        if isinstance(other, EstimationStep) or isinstance(other, SimulationStep):
            return EstimationSteps((other,) + self._steps)
        else:
            return EstimationSteps(tuple(other) + self._steps)

    def __len__(self):
        return len(self._steps)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for s1, s2 in zip(self, other):
            if s1 != s2:
                return False
        return True

    def __hash__(self):
        return hash(self._steps)

    def to_dict(self) -> dict[str, Any]:
        return {'steps': tuple(step.to_dict() for step in self._steps)}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> EstimationSteps:
        steps = []
        for sdict in d['steps']:
            if sdict['class'] == 'EstimationStep':
                s = EstimationStep.from_dict(sdict)
            else:
                s = SimulationStep.from_dict(sdict)
            steps.append(s)
        return cls(steps=tuple(steps))

    def to_dataframe(self) -> pd.DataFrame:
        """Convert to DataFrame

        Use this to create an overview of all estimation steps

        Returns
        -------
        pd.DataFrame
            DataFrame overview

        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.estimation_steps.to_dataframe()   # doctest: +ELLIPSIS
        method  interaction       parameter_uncertainty_method  ...  auto keep_every_nth_iter  tool_options
        0   FOCE         True  SANDWICH  ...  None                None            {}
        """
        steps = [s for s in self._steps if isinstance(s, EstimationStep)]
        method = [s.method for s in steps]
        interaction = [s.interaction for s in steps]
        parameter_uncertainty_method = [s.parameter_uncertainty_method for s in steps]
        evaluation = [s.evaluation for s in steps]
        maximum_evaluations = [s.maximum_evaluations for s in steps]
        laplace = [s.laplace for s in steps]
        isample = [s.isample for s in steps]
        niter = [s.niter for s in steps]
        auto = [s.auto for s in steps]
        keep_every_nth_iter = [s.keep_every_nth_iter for s in steps]
        tool_options = [dict(s.tool_options) if s.tool_options else dict() for s in steps]
        df = pd.DataFrame(
            {
                'method': method,
                'interaction': interaction,
                'parameter_uncertainty_method': parameter_uncertainty_method,
                'evaluation': evaluation,
                'maximum_evaluations': maximum_evaluations,
                'laplace': laplace,
                'isample': isample,
                'niter': niter,
                'auto': auto,
                'keep_every_nth_iter': keep_every_nth_iter,
                'tool_options': tool_options,
            }
        )
        return df

    def __repr__(self):
        if len(self) == 0:
            return "EstimationSteps()"
        return self.to_dataframe().to_string()

    def _repr_html_(self):
        if len(self) == 0:
            return "EstimationSteps()"
        else:
            return self.to_dataframe().to_html()



from pathlib import Path

from pharmpy.config import ConfigItem, Configuration
from pharmpy.internals.fs.path import normalize_user_given_path


class NONMEMConfiguration(Configuration):
    module = 'pharmpy.plugins.nonmem'  # TODO: Change default
    default_nonmem_path = ConfigItem(
        Path(''),
        'Full path to the default NONMEM installation directory',
        cls=normalize_user_given_path,
    )
    write_etas_in_abbr = ConfigItem(False, 'Whether to write etas as $ABBR records', bool)
    licfile = ConfigItem(None, 'Path to the NONMEM license file', cls=normalize_user_given_path)


conf = NONMEMConfiguration()



from pharmpy.modeling import get_bioavailability, get_lag_times


def test_get_lag_times(pheno):
    lags = get_lag_times(pheno)
    assert lags == {}


def test_get_bioavailability(pheno):
    fs = get_bioavailability(pheno)
    assert fs == {}



from typing import Iterable, Union

from pharmpy.deps import numpy as np
from pharmpy.deps.scipy import stats
from pharmpy.model import Model
from pharmpy.workflows import ModelEntry


def degrees_of_freedom(parent: Union[Model, ModelEntry], child: Union[Model, ModelEntry]) -> int:
    if isinstance(child, ModelEntry):
        child_parameters = child.model.parameters
    else:
        child_parameters = child.parameters

    if isinstance(child, ModelEntry):
        parent_parameters = parent.model.parameters
    else:
        parent_parameters = parent.parameters
    return len(child_parameters) - len(parent_parameters)


def cutoff(parent: Model, child: Model, alpha: float) -> float:
    df = degrees_of_freedom(parent, child)
    return (
        0
        if df == 0
        else float(stats.chi2.isf(q=alpha, df=df))
        if df > 0
        else -float(stats.chi2.isf(q=alpha, df=-df))
    )


def p_value(reduced: Model, extended: Model, reduced_ofv, extended_ofv) -> float:
    dofv = reduced_ofv - extended_ofv
    df = degrees_of_freedom(reduced, extended)
    return float(stats.chi2.sf(x=dofv, df=df))


def test(parent: Model, child: Model, parent_ofv, child_ofv, alpha: float) -> bool:
    dofv = parent_ofv - child_ofv
    return dofv >= cutoff(parent, child, alpha)


def best_of_two(parent: Model, child: Model, parent_ofv, child_ofv, alpha: float) -> Model:
    return child if test(parent, child, parent_ofv, child_ofv, alpha) else parent


def best_of_many(
    parent: Model, models: Iterable[Model], parent_ofv, model_ofvs, alpha: float
) -> Model:
    # NOTE: numpy.nanargmin ignores NaN values and raises a ValueError when all
    # values are NaN.
    # See https://numpy.org/doc/stable/reference/generated/numpy.nanargmin.html
    try:
        best_index = np.nanargmin(model_ofvs)
    except ValueError:
        return parent
    return best_of_two(parent, list(models)[best_index], parent_ofv, model_ofvs[best_index], alpha)



import pytest


@pytest.mark.parametrize(
    "buf,expected",
    [
        ('$MODEL NCOMPARTMENTS=23', 23),
        ('$MODEL NCOMP=18', 18),
        ('$MODEL', None),
        ('$MODEL NCOMPS=2', 2),
        ('$MODEL NCM=4', 4),
    ],
)
def test_ncomps(parser, buf, expected):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert rec.ncomps == expected


@pytest.mark.parametrize(
    "buf,add,kwargs,expected",
    [
        ('$MODEL NCOMPARTMENTS=23', 'CENTRAL', {}, '$MODEL NCOMPARTMENTS=23 COMPARTMENT=(CENTRAL)'),
        (
            '$MODEL NCOMPARTMENTS=23\n',
            'CENTRAL',
            {},
            '$MODEL NCOMPARTMENTS=23 COMPARTMENT=(CENTRAL)\n',
        ),
        (
            '$MODEL NCOM=1',
            'CENTRAL',
            {'dosing': True},
            '$MODEL NCOM=1 COMPARTMENT=(CENTRAL DEFDOSE)',
        ),
        ('$MODEL ', 'CENTRAL', {}, '$MODEL COMPARTMENT=(CENTRAL)'),
    ],
)
def test_add_compartment(parser, buf, add, kwargs, expected):
    recs = parser.parse(buf)
    rec = recs.records[0]
    newrec = rec.add_compartment(add, **kwargs)
    assert str(newrec) == expected


@pytest.mark.parametrize(
    "buf,results",
    [
        ('$MODEL COMP=1', [('1', [])]),
        ('$MODEL  COMP=(1) COMPARTMENT=(2)', [('1', []), ('2', [])]),
        (
            '$MODEL  COMP=(DEPOT DEFDOSE) COMPARTMENT=(CENTRAL DEFOBS)',
            [('DEPOT', ['DEFDOSE']), ('CENTRAL', ['DEFOBSERVATION'])],
        ),
        (
            '$MODEL  COMP=(DEFDOSE DEPOT) COMPARTMENT=(DEFOBS CENTRAL)',
            [('DEPOT', ['DEFDOSE']), ('CENTRAL', ['DEFOBSERVATION'])],
        ),
        (
            '$MODEL  COMP=(DEFDOSE) COMPARTMENT=(DEFOBS CENTRAL)',
            [('COMP1', ['DEFDOSE']), ('CENTRAL', ['DEFOBSERVATION'])],
        ),
    ],
)
def test_compartments(parser, buf, results):
    rec = parser.parse(buf).records[0]
    assert list(rec.compartments()) == results


@pytest.mark.parametrize(
    "buf,name,results",
    [
        ('$MODEL COMP=CENTRAL', 'CENTRAL', 1),
        ('$MODEL COMP=DEPOT COMP=CENTRAL', 'CENTRAL', 2),
        ('$MODEL COMP=DEPOT COMP=CENTRAL', 'DEPOT', 1),
        ('$MODEL COMP=DEPOT COMP=CENTRAL', 'NOTHING', None),
    ],
)
def test_get_compartment_number(parser, buf, name, results):
    rec = parser.parse(buf).records[0]
    assert rec.get_compartment_number(name) == results



from typing import Callable, Iterable, Set, TypeVar

T = TypeVar('T')


def reachable_from(start_nodes: Set[T], neighbors: Callable[[T], Iterable[T]]) -> Set[T]:
    queue = list(start_nodes)
    closure = set(start_nodes)
    while queue:
        u = queue.pop()
        n = neighbors(u)
        for v in n:
            if v not in closure:
                queue.append(v)
                closure.add(v)

    return closure



from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING, Optional, Tuple

from pharmpy.deps import sympy
from pharmpy.model import (
    Assignment,
    Bolus,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    DataInfo,
    Infusion,
    ModelSyntaxError,
    output,
    to_compartmental_system,
)

if TYPE_CHECKING:
    from .model import Model

from .nmtran_parser import NMTranControlStream


def compartmental_model(model: Model, advan: str, trans, des=None):
    di = model.datainfo
    control_stream = model.internals.control_stream
    return _compartmental_model(di, model.dataset, control_stream, advan, trans, des)


def _compartmental_model(
    di: DataInfo,
    dataset,
    control_stream: NMTranControlStream,
    advan: str,
    trans,
    des=None,
):
    if advan == 'ADVAN1':
        # FIXME: Require multiple doses per comp before IV+ORAL
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        cb.add_compartment(central)
        cb.add_flow(central, output, _advan1and2_trans(trans))
        ass = _f_link_assignment(control_stream, central, 1)
        comp_map = {'CENTRAL': 1, 'OUTPUT': 2}
    elif advan == 'ADVAN2':
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        depot = Compartment.create(
            'DEPOT',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=2),
            lag_time=_get_alag(control_stream, 2),
            bioavailability=_get_bioavailability(control_stream, 2),
        )
        cb.add_compartment(depot)
        cb.add_compartment(central)
        cb.add_flow(central, output, _advan1and2_trans(trans))
        cb.add_flow(depot, central, sympy.Symbol('KA'))
        ass = _f_link_assignment(control_stream, central, 2)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'OUTPUT': 3}
    elif advan == 'ADVAN3':
        # FIXME: Multiple doses per compartment IV+ORAL
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        peripheral = Compartment.create(
            'PERIPHERAL',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 2),
            bioavailability=_get_bioavailability(control_stream, 2),
        )
        cb.add_compartment(central)
        cb.add_compartment(peripheral)
        k, k12, k21 = _advan3_trans(trans)
        cb.add_flow(central, output, k)
        cb.add_flow(central, peripheral, k12)
        cb.add_flow(peripheral, central, k21)
        ass = _f_link_assignment(control_stream, central, 1)
        comp_map = {'CENTRAL': 1, 'PERIPHERAL': 2, 'OUTPUT': 3}
    elif advan == 'ADVAN4':
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        depot = Compartment.create(
            'DEPOT',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=2),
            lag_time=_get_alag(control_stream, 2),
            bioavailability=_get_bioavailability(control_stream, 2),
        )
        peripheral = Compartment.create(
            'PERIPHERAL',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 3),
            bioavailability=_get_bioavailability(control_stream, 3),
        )
        cb.add_compartment(depot)
        cb.add_compartment(central)
        cb.add_compartment(peripheral)
        k, k23, k32, ka = _advan4_trans(trans)
        cb.add_flow(depot, central, ka)
        cb.add_flow(central, output, k)
        cb.add_flow(central, peripheral, k23)
        cb.add_flow(peripheral, central, k32)
        ass = _f_link_assignment(control_stream, central, 2)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'PERIPHERAL': 3, 'OUTPUT': 4}
    elif advan == 'ADVAN5' or advan == 'ADVAN7':
        cb = CompartmentalSystemBuilder()
        modrec = control_stream.get_records('MODEL')[0]
        defobs: Optional[Tuple[str, int]] = None
        defdose: Optional[Tuple[str, int]] = None
        defcentral: Optional[Tuple[str, int]] = None
        defdepot: Optional[Tuple[str, int]] = None
        deffirst_dose: Optional[Tuple[str, int]] = None
        compartments = []
        comp_names = []
        for i, (name, opts) in enumerate(modrec.compartments(), 1):
            if 'DEFOBSERVATION' in opts:
                defobs = (name, i)
            if 'DEFDOSE' in opts:
                defdose = (name, i)
            if name == 'CENTRAL':
                defcentral = (name, i)
            elif name == 'DEPOT':
                defdepot = (name, i)
            if deffirst_dose is None and 'NODOSE' not in opts:
                deffirst_dose = (name, i)
            comp_names.append(name)

        comp_map = {name: i for i, name in enumerate(comp_names, 1)}

        if defobs is None:
            if defcentral is None:
                defobs = (comp_names[0], 1)
            else:
                defobs = defcentral

        if defdose is None:
            if defdepot is not None:
                defdose = defdepot
            elif deffirst_dose is not None:
                defdose = deffirst_dose
            else:
                raise ModelSyntaxError('Dosing compartment is unknown')

        doses = dosing(di, dataset, defdose[1])
        obscomp = None
        for i, name in enumerate(comp_names, start=1):
            curdose = find_dose(doses, i)
            comp = Compartment.create(
                name,
                doses=curdose,
                lag_time=_get_alag(control_stream, i),
                bioavailability=_get_bioavailability(control_stream, i),
            )
            cb.add_compartment(comp)
            compartments.append(comp)
            if name == defobs[0]:
                obscomp = comp
        compartments.append(output)

        assert obscomp is not None
        for from_n, to_n, rate in _find_rates(control_stream, len(compartments)):
            cb.add_flow(compartments[from_n - 1], compartments[to_n - 1], rate)
        ass = _f_link_assignment(control_stream, obscomp, defobs[1])
    elif advan == 'ADVAN10':
        # FIXME: Multiple doses per compartment needed
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        cb.add_compartment(central)
        vm = sympy.Symbol('VM')
        km = sympy.Symbol('KM')
        t = sympy.Symbol('t')
        cb.add_flow(central, output, vm / (km + sympy.Function(central.amount.name)(t)))
        ass = _f_link_assignment(control_stream, central, 1)
        comp_map = {'CENTRAL': 1, 'OUTPUT': 2}
    elif advan == 'ADVAN11':
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        per1 = Compartment.create(
            'PERIPHERAL1',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 2),
            bioavailability=_get_bioavailability(control_stream, 2),
        )
        per2 = Compartment.create(
            'PERIPHERAL2',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 3),
            bioavailability=_get_bioavailability(control_stream, 3),
        )
        cb.add_compartment(central)
        cb.add_compartment(per1)
        cb.add_compartment(per2)
        k, k12, k21, k13, k31 = _advan11_trans(trans)
        cb.add_flow(central, output, k)
        cb.add_flow(central, per1, k12)
        cb.add_flow(per1, central, k21)
        cb.add_flow(central, per2, k13)
        cb.add_flow(per2, central, k31)
        ass = _f_link_assignment(control_stream, central, 1)
        comp_map = {'CENTRAL': 1, 'PERIPHERAL1': 2, 'PERIPHERAL2': 3, 'OUTPUT': 4}
    elif advan == 'ADVAN12':
        doses = dosing(di, dataset, 1)
        cb = CompartmentalSystemBuilder()
        depot = Compartment.create(
            'DEPOT',
            doses=find_dose(doses, comp_number=1),
            lag_time=_get_alag(control_stream, 1),
            bioavailability=_get_bioavailability(control_stream, 1),
        )
        central = Compartment.create(
            'CENTRAL',
            doses=find_dose(doses, comp_number=2),
            lag_time=_get_alag(control_stream, 2),
            bioavailability=_get_bioavailability(control_stream, 2),
        )
        per1 = Compartment.create(
            'PERIPHERAL1',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 3),
            bioavailability=_get_bioavailability(control_stream, 3),
        )
        per2 = Compartment.create(
            'PERIPHERAL2',
            doses=tuple(),
            lag_time=_get_alag(control_stream, 4),
            bioavailability=_get_bioavailability(control_stream, 4),
        )
        cb.add_compartment(depot)
        cb.add_compartment(central)
        cb.add_compartment(per1)
        cb.add_compartment(per2)
        k, k23, k32, k24, k42, ka = _advan12_trans(trans)
        cb.add_flow(depot, central, ka)
        cb.add_flow(central, output, k)
        cb.add_flow(central, per1, k23)
        cb.add_flow(per1, central, k32)
        cb.add_flow(central, per2, k24)
        cb.add_flow(per2, central, k42)
        ass = _f_link_assignment(control_stream, central, 2)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'PERIPHERAL1': 3, 'PERIPHERAL2': 4, 'OUTPUT': 5}
    elif des:
        # FIXME: Add dose based on presence of CMT column

        rec_model = control_stream.get_records('MODEL')[0]

        subs_dict, comp_names = {}, {}
        comps = [c for c, _ in rec_model.compartments()]
        func_to_name = {}

        t = sympy.Symbol('t')
        subs_dict['T'] = t
        for i, c in enumerate(comps, 1):
            a = sympy.Function(f'A_{c}')
            subs_dict[f'DADT({i})'] = sympy.Derivative(a(t))
            subs_dict[f'DADT ({i})'] = sympy.Derivative(a(t))
            subs_dict[f'A({i})'] = a(t)
            comp_names[f'A({i})'] = a
            func_to_name[a] = c

        sset = des.statements.subs(subs_dict)

        eqs = [sympy.Eq(s.symbol, s.expression) for s in sset if not s.symbol.is_Symbol]

        cs = to_compartmental_system(func_to_name, eqs)
        cb = CompartmentalSystemBuilder(cs)
        doses = dosing(di, dataset, 1)
        comp_map = {}
        for i, comp_name in enumerate(comps, start=1):
            comp = cs.find_compartment(comp_name)
            comp_map[comp_name] = i
            if comp is None:  # Compartments can be in $MODEL but not used in $DES
                continue
            cb.set_dose(comp, find_dose(doses, i))
            comp = cb.find_compartment(comp_name)
            f = _get_bioavailability(control_stream, i)
            cb.set_bioavailability(comp, f)
            comp = cb.find_compartment(comp_name)
            alag = _get_alag(control_stream, i)
            cb.set_lag_time(comp, alag)

        # Search for DEFOBSERVATION, default to first
        it = iter(rec_model.compartments())
        defobs = (next(it)[0], 1)
        for i, (name, opts) in enumerate(it, start=2):
            if 'DEFOBSERVATION' in opts:
                defobs = (name, i)

        ass = _f_link_assignment(control_stream, sympy.Symbol(f'A_{defobs[0]}'), defobs[1])
    else:
        return None
    return CompartmentalSystem(cb), ass, comp_map


def des_assign_statements(
    control_stream: NMTranControlStream,
    des=None,
):
    if des:
        rec_model = control_stream.get_records('MODEL')[0]

        subs_dict, comp_names = {}, {}
        comps = [c for c, _ in rec_model.compartments()]
        func_to_name = {}
        t = sympy.Symbol('t')
        for i, c in enumerate(comps, 1):
            a = sympy.Function(f'A_{c}')
            subs_dict[f'DADT({i})'] = sympy.Derivative(a(t))
            subs_dict[f'DADT ({i})'] = sympy.Derivative(a(t))
            subs_dict[f'A({i})'] = a(t)
            comp_names[f'A({i})'] = a
            func_to_name[a] = c

        sset = des.statements.subs(subs_dict)

        statements = [sympy.Eq(s.symbol, s.expression) for s in sset if s.symbol.is_Symbol]
        if len(statements) == 0:
            statements = None
        return statements


def _f_link_assignment(control_stream: NMTranControlStream, compartment: Compartment, compno: int):
    f = sympy.Symbol('F')
    try:
        fexpr = compartment.amount
    except AttributeError:
        fexpr = compartment
    fexpr = sympy.Function(fexpr.name)(sympy.Symbol('t'))
    pkrec = control_stream.get_records('PK')[0]
    scaling = f'S{compno}'
    if pkrec.statements.find_assignment(scaling):
        fexpr = fexpr / sympy.Symbol(scaling)
    ass = Assignment(f, fexpr)
    return ass


def _find_rates(control_stream: NMTranControlStream, ncomps: int):
    pkrec = control_stream.get_records('PK')[0]
    for stat in pkrec.statements:
        if hasattr(stat, 'symbol'):
            name = stat.symbol.name
            m = re.match(r'^K(\d+)(T\d+)?$', name)
            if m:
                if m.group(2):
                    from_n = int(m.group(1))
                    to_n = int(m.group(2)[1:])
                else:
                    n = m.group(1)
                    if len(n) == 2:
                        from_n = int(n[0])
                        to_n = int(n[1])
                    elif len(n) == 3:
                        f1 = int(n[0])
                        t1 = int(n[1:])
                        f2 = int(n[:2])
                        t2 = int(n[2:])
                        q1 = f1 <= ncomps and t1 <= ncomps and t1 != 0
                        q2 = f2 <= ncomps and t2 <= ncomps
                        if q1 and q2:
                            raise ModelSyntaxError(
                                f'Rate parameter {n} is ambiguous. Use the KiTj notation.'
                            )
                        if q1:
                            from_n = f1
                            to_n = t1
                        elif q2:
                            from_n = f2
                            to_n = t2
                        else:
                            # Too large to or from compartment index. What would NONMEM do?
                            # Could also be too large later
                            continue
                    elif len(n) == 4:
                        from_n = int(n[:2])
                        to_n = int(n[2:])
                    else:
                        raise ValueError(f'Cannot handle {n}.')

                if to_n == 0:
                    to_n = ncomps

                yield from_n, to_n, sympy.Symbol(name)


def _advan1and2_trans(trans: str):
    if trans == 'TRANS2':
        return sympy.Symbol('CL') / sympy.Symbol('V')
    else:  # TRANS1 which is also the default
        return sympy.Symbol('K')


def _advan3_trans(trans: str):
    if trans == 'TRANS3':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V'),
            sympy.Symbol('Q') / sympy.Symbol('V'),
            sympy.Symbol('Q') / (sympy.Symbol('VSS') - sympy.Symbol('V')),
        )
    elif trans == 'TRANS4':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V1'),
            sympy.Symbol('Q') / sympy.Symbol('V1'),
            sympy.Symbol('Q') / sympy.Symbol('V2'),
        )
    elif trans == 'TRANS5':
        return (
            sympy.Symbol('ALPHA') * sympy.Symbol('BETA') / sympy.Symbol('K21'),
            sympy.Symbol('ALPHA') + sympy.Symbol('BETA') - sympy.Symbol('K21') - sympy.Symbol('K'),
            (sympy.Symbol('AOB') * sympy.Symbol('BETA') + sympy.Symbol('ALPHA'))
            / (sympy.Symbol('AOB') + 1),
        )
    elif trans == 'TRANS6':
        return (
            sympy.Symbol('ALPHA') * sympy.Symbol('BETA') / sympy.Symbol('K21'),
            sympy.Symbol('ALPHA') + sympy.Symbol('BETA') - sympy.Symbol('K21') - sympy.Symbol('K'),
            sympy.Symbol('K21'),
        )
    else:
        return (sympy.Symbol('K'), sympy.Symbol('K12'), sympy.Symbol('K21'))


def _advan4_trans(trans: str):
    if trans == 'TRANS3':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V'),
            sympy.Symbol('Q') / sympy.Symbol('V'),
            sympy.Symbol('Q') / (sympy.Symbol('VSS') - sympy.Symbol('V')),
            sympy.Symbol('KA'),
        )
    elif trans == 'TRANS4':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V2'),
            sympy.Symbol('Q') / sympy.Symbol('V2'),
            sympy.Symbol('Q') / sympy.Symbol('V3'),
            sympy.Symbol('KA'),
        )
    elif trans == 'TRANS5':
        return (
            sympy.Symbol('ALPHA') * sympy.Symbol('BETA') / sympy.Symbol('K32'),
            sympy.Symbol('ALPHA') + sympy.Symbol('BETA') - sympy.Symbol('K32') - sympy.Symbol('K'),
            (sympy.Symbol('AOB') * sympy.Symbol('BETA') + sympy.Symbol('ALPHA'))
            / (sympy.Symbol('AOB') + 1),
            sympy.Symbol('KA'),
        )
    elif trans == 'TRANS6':
        return (
            sympy.Symbol('ALPHA') * sympy.Symbol('BETA') / sympy.Symbol('K32'),
            sympy.Symbol('ALPHA') + sympy.Symbol('BETA') - sympy.Symbol('K32') - sympy.Symbol('K'),
            sympy.Symbol('K32'),
            sympy.Symbol('KA'),
        )
    else:
        return (sympy.Symbol('K'), sympy.Symbol('K23'), sympy.Symbol('K32'), sympy.Symbol('KA'))


def _advan11_trans(trans: str):
    if trans == 'TRANS4':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V1'),
            sympy.Symbol('Q2') / sympy.Symbol('V1'),
            sympy.Symbol('Q2') / sympy.Symbol('V2'),
            sympy.Symbol('Q3') / sympy.Symbol('V1'),
            sympy.Symbol('Q3') / sympy.Symbol('V3'),
        )
    elif trans == 'TRANS6':
        return (
            sympy.Symbol('ALPHA')
            * sympy.Symbol('BETA')
            * sympy.Symbol('GAMMA')
            / (sympy.Symbol('K21') * sympy.Symbol('K31')),
            sympy.Symbol('ALPHA')
            + sympy.Symbol('BETA')
            + sympy.Symbol('GAMMA')
            - sympy.Symbol('K')
            - sympy.Symbol('K13')
            - sympy.Symbol('K21')
            - sympy.Symbol('K31'),
            sympy.Symbol('K21'),
            (
                sympy.Symbol('ALPHA') * sympy.Symbol('BETA')
                + sympy.Symbol('ALPHA') * sympy.Symbol('GAMMA')
                + sympy.Symbol('BETA') * sympy.Symbol('GAMMA')
                + sympy.Symbol('K31') * sympy.Symbol('K31')
                - sympy.Symbol('K31')
                * (sympy.Symbol('ALPHA') + sympy.Symbol('BETA') + sympy.Symbol('GAMMA'))
                - sympy.Symbol('K') * sympy.Symbol('K21')
            )
            / (sympy.Symbol('K21') - sympy.Symbol('K31')),
            sympy.Symbol('K31'),
        )
    else:
        return (
            sympy.Symbol('K'),
            sympy.Symbol('K12'),
            sympy.Symbol('K21'),
            sympy.Symbol('K13'),
            sympy.Symbol('K31'),
        )


def _advan12_trans(trans: str):
    if trans == 'TRANS4':
        return (
            sympy.Symbol('CL') / sympy.Symbol('V2'),
            sympy.Symbol('Q3') / sympy.Symbol('V2'),
            sympy.Symbol('Q3') / sympy.Symbol('V3'),
            sympy.Symbol('Q4') / sympy.Symbol('V2'),
            sympy.Symbol('Q4') / sympy.Symbol('V4'),
            sympy.Symbol('KA'),
        )
    elif trans == 'TRANS6':
        return (
            sympy.Symbol('ALPHA')
            * sympy.Symbol('BETA')
            * sympy.Symbol('GAMMA')
            / (sympy.Symbol('K32') * sympy.Symbol('K42')),
            sympy.Symbol('ALPHA')
            + sympy.Symbol('BETA')
            + sympy.Symbol('GAMMA')
            - sympy.Symbol('K')
            - sympy.Symbol('K24')
            - sympy.Symbol('K32')
            - sympy.Symbol('K42'),
            sympy.Symbol('K32'),
            (
                sympy.Symbol('ALPHA') * sympy.Symbol('BETA')
                + sympy.Symbol('ALPHA') * sympy.Symbol('GAMMA')
                + sympy.Symbol('BETA') * sympy.Symbol('GAMMA')
                + sympy.Symbol('K42') * sympy.Symbol('K42')
                - sympy.Symbol('K42')
                * (sympy.Symbol('ALPHA') + sympy.Symbol('BETA') + sympy.Symbol('GAMMA'))
                - sympy.Symbol('K') * sympy.Symbol('K32')
            )
            / (sympy.Symbol('K32') - sympy.Symbol('K42')),
            sympy.Symbol('K42'),
            sympy.Symbol('KA'),
        )
    else:
        return (
            sympy.Symbol('K'),
            sympy.Symbol('K23'),
            sympy.Symbol('K32'),
            sympy.Symbol('K24'),
            sympy.Symbol('K42'),
            sympy.Symbol('KA'),
        )


def dosing(di: DataInfo, dataset, dose_comp: int):
    # Only check doses
    if dataset is not None:
        dataset = dataset[dataset['AMT'] != 0]

    cmt_loop = False
    admid_name = None
    return_dose = False
    if dataset is None:
        return_dose = True
    elif 'admid' in di.types:
        admid_name = di.typeix["admid"][0].name
        if 'CMT' in di.names and not di['CMT'].drop:
            if len(dataset['CMT']) == 1:
                warnings.warn("CMT column present with only one value")
            cmt_loop = True
    elif 'CMT' in di.names and not di['CMT'].drop:
        if len(dataset['CMT']) == 1:
            warnings.warn("CMT column present with only one value")
            return_dose = True
        else:
            cmt_loop = True
    else:
        return_dose = True

    if return_dose:
        return (
            {
                'comp_number': dose_comp,
                'dose': _dosing(di, dataset, dose_comp),
                'admid': None,
            },
        )

    doses = tuple()
    if admid_name is not None and cmt_loop:
        for comp_number in dataset['CMT'].unique():
            cmt_dataset = dataset[dataset['CMT'] == comp_number]
            for admid in cmt_dataset[admid_name].unique():
                doses += (
                    {
                        'comp_number': comp_number,
                        'dose': _dosing(di, dataset.loc[dataset[admid_name] == admid], comp_number),
                        'admid': admid,
                    },
                )
    elif admid_name is not None:
        for admid in dataset[admid_name].unique():
            doses += (
                {
                    'comp_number': dose_comp,
                    'dose': _dosing(di, dataset.loc[dataset[admid_name] == admid], dose_comp),
                    'admid': admid,
                },
            )
    elif cmt_loop:
        for comp_number in dataset['CMT'].unique():
            cmt_dataset = dataset[dataset['CMT'] == comp_number]
            doses += (
                {
                    'comp_number': comp_number,
                    'dose': _dosing(di, cmt_dataset, comp_number),
                    'admid': None,
                },
            )

    return doses


def _dosing(di, dataset, dose_comp):
    if 'RATE' not in di.names or di['RATE'].drop:
        return Bolus(sympy.Symbol('AMT'))

    df = dataset

    if df is None:
        return Bolus(sympy.Symbol('AMT'))
    elif (df['RATE'] == 0).all():
        return Bolus(sympy.Symbol('AMT'))
    elif (df['RATE'] == -1).any():
        return Infusion(sympy.Symbol('AMT'), rate=sympy.Symbol(f'R{dose_comp}'))
    elif (df['RATE'] == -2).any():
        return Infusion(sympy.Symbol('AMT'), duration=sympy.Symbol(f'D{dose_comp}'))
    else:
        return Infusion(sympy.Symbol('AMT'), rate=sympy.Symbol('RATE'))


def find_dose(doses, comp_number, admid=1):
    comp_doses = tuple()
    for dose in doses:
        if dose['comp_number'] == comp_number:
            comp_dose = dose['dose']
            if dose['admid'] is not None:
                admid = dose['admid']
            comp_doses += (comp_dose.replace(admid=admid),)
    return comp_doses


def _get_alag(control_stream: NMTranControlStream, n: int):
    """Check if ALAGn is defined in model and return it else return 0"""
    alag = f'ALAG{n}'
    pkrec = control_stream.get_records('PK')[0]
    if pkrec.statements.find_assignment(alag):
        return sympy.Symbol(alag)
    else:
        return sympy.Integer(0)


def _get_bioavailability(control_stream: NMTranControlStream, n: int):
    """Check if Fn is defined in model and return it else return 0"""
    fn = f'F{n}'
    pkrec = control_stream.get_records('PK')[0]
    if pkrec.statements.find_assignment(fn):
        return sympy.Symbol(fn)
    else:
        return sympy.Integer(1)



from functools import partial
from typing import Iterable

from pharmpy.model import Model
from pharmpy.modeling import add_metabolite

from ..statement.feature.metabolite import Metabolite
from ..statement.feature.symbols import Name, Wildcard
from ..statement.statement import Statement
from .feature import Feature


def features(model: Model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, Metabolite):
            modes = (
                [Name('BASIC'), Name('PSC')]
                if isinstance(statement.modes, Wildcard)
                else statement.modes
            )
            for mode in modes:
                yield ('METABOLITE', mode.name), partial(
                    add_metabolite, presystemic=True if mode.name == "PSC" else False
                )



from typing import Callable, Hashable, Optional, Tuple

from pharmpy.model import Model

FeatureKey = Tuple[Hashable, ...]
FeatureFn = Callable[[Model], Optional[Model]]
Feature = Tuple[FeatureKey, FeatureFn]



"""
Pretty printing of tree-structures. Inspired by the 'tree' command (Steve
Baker). In development.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Generic, Optional, Tuple, TypeVar, Union


class BranchStyle(Enum):
    SIMPLE = 1
    INLINE = 2
    BUSY = 3


Indent = namedtuple('Indent', ['header', 'node', 'fork'])

CharSet = namedtuple('CharSet', ['fork', 'horiz', 'vert', 'inline', 'lfork', 'end'])


@dataclass(frozen=True)
class NodeStyle:
    branch: BranchStyle = BranchStyle['SIMPLE']
    indent: Indent = Indent(header=0, node=1, fork=1)
    char: CharSet = CharSet(fork='├', horiz='│', vert='─', inline='┌', lfork='└', end=' ')

    def __post_init__(self):
        log = logging.getLogger(__name__ + '.' + self.__class__.__name__)
        if self.indent.header > self.indent.node:
            log.warning("Indent of header > indent of node. Might break assumptions.")
        for x in self.char:
            if len(x) != 1:
                raise ValueError(f'len({x}) != 1 (got {len(x)})')


T = TypeVar('T')


class Node(Generic[T]):
    def __init__(
        self,
        obj: T,
        cls_str: Union[str, Callable[[T], str]] = '__str__',
        style: Optional[NodeStyle] = None,
        children: Tuple[Node, ...] = (),
    ):
        self._obj = obj
        self._cls_str = cls_str
        self._style = style
        self._parent = None
        self._children = children
        # NOTE: This links the children to the new parent and forces to build
        # the tree from bottom to top
        for child in children:
            child._parent = self

    @property
    def style(self):
        if self._style:
            return self._style
        elif self._parent:
            return self._parent.style
        else:
            return NodeStyle()

    @style.setter
    def style(self, **kwargs):
        self._style = NodeStyle(**kwargs)

    @property
    def lines(self):
        if isinstance(self._cls_str, str):
            attr = getattr(self._obj, self._cls_str)
            try:
                out = str(attr())
            except TypeError:
                out = str(attr)
        else:
            out = str(self._cls_str(self._obj))

        return out.splitlines() if isinstance(out, str) else out

    def _prefix(self, lines, early_branch, ind_used):
        if self.style.branch != BranchStyle.SIMPLE:
            raise NotImplementedError('only BranchStyle.SIMPLE suppported')
        ind = self.style.indent
        char = self.style.char
        if early_branch:
            ch_first, ch_after = (char.fork, char.horiz)
        else:
            ch_first, ch_after = (char.lfork, ' ')

        vnew = char.vert * ind.node
        vext = char.vert * ind_used.header
        for i, line in enumerate(lines):
            if i == 0:
                lines[i] = ' ' * ind.fork + ch_first + vnew + vext + char.end + line[len(vext) :]
                continue
            lines[i] = ' ' * ind.fork + ch_after + ' ' * ind.node + line

        return lines

    def __str__(self):
        ind = self.style.indent
        lines = self.lines
        lines[0] = ' ' * ind.header + lines[0]

        idx_last = len(self._children) - 1
        for idx, node in enumerate(self._children):
            nlines = str(node).splitlines()
            if not nlines:
                continue
            nlines = self._prefix(nlines, idx < idx_last, node.style.indent)
            lines += nlines

        return '\n'.join(lines)

    def __repr__(self):
        head = f'{self.__class__.__name__}({self._obj.__class__.__name__}) {repr(self._obj)}'
        lines = [' ' + line for child in self._children for line in repr(child).splitlines()]
        return '\n'.join([head] + lines)


def from_ast(ast_node):
    from ast import iter_child_nodes

    children = tuple(map(from_ast, iter_child_nodes(ast_node)))
    root = Node(ast_node, children=children)
    return root


if __name__ == '__main__':
    with open(__file__, 'r') as source:
        from ast import parse

        root = parse(source.read())

    root = from_ast(root)
    print('=' * 40, 'repr(...)', '=' * 40)
    print(repr(root))
    print()
    print('=' * 40, 'str(...)', '=' * 40)
    print(str(root))



from pharmpy.internals.module.lazy import LazyImport

str = LazyImport('str', globals(), 'sympy.printing.str')
fortran = LazyImport('fortran', globals(), 'sympy.printing.fortran')
codeprinter = LazyImport('codeprinter', globals(), 'sympy.printing.codeprinter')
